# -*- coding: utf-8 -*-
""" Модуль отвечающий за основной расчет схемы цепи, модуль также описывает класс, хранящий основные настройки.
"""

# TODO: Попробовать перейти на scipy.optimize.

import cmath
import math
from numpy import linalg, zeros
import classes
import language
import prog_settings


class Settings(object):

    def __init__(self):
        #стандартные значения основных переменных.
        #показатель сходимости
        self.delta_norm = 1E-3
        #максимальное количество итераций.
        self.max_shod = 2000
        #показатель округления.
        self.round_n = 3
        #разделитель десятичных дробей в csv файлах.
        self.file_point = ','


class NodeScheme(object):

    def __init__(self, settings=Settings()):
        # Входные данные.
        self.settings = settings
        # Коэффициент единицы измерения активной мощности.
        self.unit_k_active_power = 1.0
        # Коэффициент единицы измерения реактивной мощности.
        self.unit_k_reactive_power = 1.0
        # Коэффициент единицы измерения полной мощности.
        self.unit_k_full_power = 1.0
        # Коэффициент единицы измерения напряжения.
        self.unit_k_voltage = 1.0
        # Коэффициент единицы измерения тока.
        self.unit_k_current = 1.0

        # Выходные параметры.
        # Базисное напряжение.
        self.base_u = 1.0
        # Количество итераций.
        self.n_shod = 0
        # Объект с параметрами ветвей.
        self.vetvi = classes.Parameters('vetvi')
        # Объект с параметрами узлов.
        self.usli = classes.Parameters('usli')

    def read_base_voltage(self, base_voltage_input):
        # Считывание базисного напряжения
        errors = language.Errors()
        warnings = []
        if base_voltage_input:
            if isinstance(base_voltage_input, basestring):
                base_voltage_input = base_voltage_input.replace(',', '.')
            try:
                self.base_u = float(base_voltage_input) * self.unit_k_voltage
            except ValueError:
                warnings = [u'%s: %s' % (errors.messages['warning'], errors.messages['incorrect base voltage'])]
                self.base_u = 1.0
        return [], warnings, []

    # Функция чтения уже считанного файла.
    def read_scheme(self, base_u_input='1.0'):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Если нет ветвей
        if not self.vetvi.child_count():
            errs.append(u'%s: %s' % (errors.messages['error'], errors.messages['elements parameters not set']))
            return errs, warnings, messages
        # Считывание базисного напряжения
        result = self.read_base_voltage(base_u_input)
        errs += result[0]
        warnings += result[1]
        messages += result[2]

        # Проход по ветвям.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            input_parameters = vetv.input_parameters()
            if input_parameters.children_list():
                # Чтение входных параметров ветви

                def read_param_float(param, def_value, mult_value=1., err=u''):
                    if not param:
                        return def_value
                    if isinstance(param, int) or isinstance(param, float):
                        param *= mult_value
                    else:
                        param = def_value
                        if err:
                            errs.append(err)
                    return param

                # Чтение реального напряжения ветви.
                real_voltage = input_parameters.real_voltage().value()
                real_voltage = read_param_float(real_voltage, self.base_u, self.unit_k_voltage,
                                                u'%s: %s "%s"' % (errors.messages['error'],
                                                                  errors.messages['invalid actual element voltage'],
                                                                  vetv.name()))
                # Чтение реальной фазы ветви.
                real_phase = input_parameters.real_phase().value()
                real_phase = read_param_float(real_phase, 0., 1.,
                                              u'%s: %s "%s"' % (errors.messages['error'],
                                                                errors.messages['invalid actual element phase'],
                                                                vetv.name()))
                # Коэффициент приведения.
                k_p = self.base_u / cmath.rect(real_voltage, math.radians(real_phase))

                # Чтение модуля ЭДС источника в ветви.
                eds = input_parameters.eds().value()
                eds = read_param_float(eds, 0., self.unit_k_voltage,
                                       u'%s: %s "%s"' % (errors.messages['error'],
                                                         errors.messages['invalid module EFM'],
                                                         vetv.name()))

                # Чтение фазы ЭДС источника в ветви.
                eds_psi = input_parameters.eds_psi().value()
                eds_psi = read_param_float(eds_psi, 0., 1.,
                                           u'%s: %s "%s"' % (errors.messages['error'],
                                                             errors.messages['invalid phase EFM'],
                                                             vetv.name()))

                e = cmath.rect(eds, math.radians(eds_psi)) * k_p

                # Чтение модуля номинального напряжения ветви.
                voltage = input_parameters.voltage().value()
                voltage = read_param_float(voltage, 0., self.unit_k_voltage,
                                           u'%s: %s "%s"' % (errors.messages['error'],
                                                             errors.messages['invalid module nominal voltage'],
                                                             vetv.name()))

                # Чтение фазы номинального напряжения ветви.
                voltage_psi = input_parameters.voltage_psi().value()
                voltage_psi = read_param_float(voltage_psi, 0., 1.,
                                               u'%s: %s "%s"' % (errors.messages['error'],
                                                                 errors.messages['invalid phase nominal voltage'],
                                                                 vetv.name()))

                u = cmath.rect(voltage, math.radians(voltage_psi)) * k_p

                # Чтение активной мощности ветви.
                active_power = input_parameters.active_power().value()
                active_power = read_param_float(active_power, 0., self.unit_k_active_power,
                                                u'%s: %s "%s"' % (errors.messages['error'],
                                                                  errors.messages['invalid active power'],
                                                                  vetv.name()))

                # Чтение реактивной мощности ветви.
                reactive_power = input_parameters.reactive_power().value()
                reactive_power = read_param_float(reactive_power, 0., self.unit_k_reactive_power,
                                                  u'%s: %s "%s"' % (errors.messages['error'],
                                                                    errors.messages['invalid reactive power'],
                                                                    vetv.name()))

                s = active_power + reactive_power * 1j

                # Чтение модуля тока источника тока в ветви.
                current = input_parameters.current().value()
                current = read_param_float(current, 0., self.unit_k_current,
                                           u'%s: %s "%s"' % (errors.messages['error'],
                                                             errors.messages['invalid module current'],
                                                             vetv.name()))

                # Чтение фазы тока источника тока в ветви.
                current_psi = input_parameters.current_psi().value()
                current_psi = read_param_float(current_psi, 0., 1.,
                                               u'%s: %s "%s"' % (errors.messages['error'],
                                                                 errors.messages['invalid phase current'],
                                                                 vetv.name()))

                i = cmath.rect(current, math.radians(current_psi)) / k_p.conjugate()

                # Чтение активного и реактивного сопротивления ветви.
                r = input_parameters.active_resistance().value()
                x = input_parameters.reactive_resistance().value()

                # Чтение начального и конечного узлов.
                begin = input_parameters.begin().value()
                if not begin:
                    errs.append(u'%s: %s "%s"' % (errors.messages['error'], errors.messages['unknown initial node'],
                                                  vetv.name()))
                end = input_parameters.end().value()
                if not end:
                    errs.append(u'%s: %s "%s"' % (errors.messages['error'], errors.messages['unknown end node'],
                                                  vetv.name()))

                # если не указаны R и X.
                if r < 0. and x < 0.:
                    y = 1E-30
                else:
                    # если X = 0.
                    if r != 0. and r == 0.:
                        # Y=(1/R) / Kp^2.
                        y = (1 / r) / (abs(k_p) * abs(k_p))
                    # если R = 0.
                    elif r == 0. and x != 0.:
                        # Y=(1/jX) / Kp^2.
                        y = (1 / x * 1j) / (abs(k_p) * abs(k_p))
                    # если R = 0 и X = 0.
                    elif r == 0. and x == 0.:
                        y = 1E5
                    else:
                        # Y=(1/Z) / Kp^2.
                        y = (1 / (r + x * 1j)) / (abs(k_p) * abs(k_p))

                # Проверки по типам ветвей.
                vetv_type = input_parameters.vetv_type().value()
                if not vetv_type:
                    errs.append(u'%s: %s "%s"' % (errors.messages['error'],
                                                  errors.messages['not set type element'], vetv.name()))
                if vetv_type and begin and end:
                    # Если заземление.
                    if vetv_type == 'ground':
                        y = 1E30
                    # Если нагрузка.
                    if vetv_type in prog_settings.NAGR:
                        i = 0. + 0.j
                        # если S не равна нулю.
                        if s != 0.:
                            if u:
                                # Y = sopr(S / U) / U.
                                y = (s / u).conjugate() / u
                            else:
                                y = 1E-30
                                errs.append(u'%s: %s "%s"' % (errors.messages['error'],
                                                              errors.messages['not set nominal voltage'],
                                                              vetv.name()))
                        # если S равна нулю.
                        elif s == 0.:
                            y = 1E-30
                    # Если нагрузка с ЭДС.
                    elif vetv_type in prog_settings.NAGR_EDS:
                        i = 0. + 0.j
                        # Если Y не равно нулю.
                        if y != 0.:
                            # если S не равна нулю.
                            if s != 0.:
                                if voltage != 0.:
                                    # E = Unom - sopr(S)/(sopr(Unom)*Y).
                                    e = u - s.conjugate() / (u * y)
                                else:
                                    y = 1E-30
                                    e = 0. + 0.j
                                    errs.append(u'%s: %s "%s"' % (errors.messages['error'],
                                                                  errors.messages['not set nominal voltage'],
                                                                  vetv.name()))
                            # если S равна нулю.
                            elif s == 0.:
                                y = 1E-30
                                e = 0. + 0.j
                        else:
                            y = 1E-30
                            e = 0. + 0.j
                            errs.append(u'%s: %s "%s"' % (errors.messages['error'],
                                                          errors.messages['not set resistance'], vetv.name()))
                    # Если ЭДС.
                    elif vetv_type in prog_settings.EDS:
                        i = 0. + 0.j
                        if vetv_type not in ('EDS_PQ_perem', 'EDS_PQ_post'):
                            if not u:
                                errs.append(u'%s: %s "%s"' % (errors.messages['error'],
                                                              errors.messages['not set nominal voltage'],
                                                              vetv.name()))
                            e = u
                        else:
                            if not e and not u:
                                errs.append(u'%s: %s "%s"' % (errors.messages['error'],
                                                              errors.messages['not set EFM'],
                                                              vetv.name()))
                            elif not e and u:
                                e = u
                        # Если R или X равны нулю.
                        if y == 1E30:
                            errs.append(u'%s: %s "%s"' % (errors.messages['error'],
                                                          errors.messages['not set resistance'], vetv.name()))
                            y = 1E5
                    # Если источник тока.
                    elif vetv_type in prog_settings.II:
                        e = 0. + 0.j
                        if vetv_type == 'ist_toka_U_post':
                            i = 0. + 0.j
                        else:
                            if not i:
                                errs.append(u'%s: %s "%s"' % (errors.messages['error'],
                                                              errors.messages['not set current'], vetv.name()))
                        y = 0.

                    # Удаление старых выходных параметров.
                    output_parameters = vetv.output_parameters()
                    output_parameters.remove_children()
                    # Запись считанных данных.
                    output_parameters.e().set_value(e)
                    output_parameters.u().set_value(u)
                    output_parameters.y().set_value(y)
                    output_parameters.j().set_value(i)
                    output_parameters.s().set_value(s)
                    output_parameters.k_p().set_value(k_p)
                    output_parameters.begin().set_value(begin)
                    output_parameters.end().set_value(end)
            else:
                errs.append(u'%s: %s "%s"' % (errors.messages['error'],
                                              errors.messages['not set parameters'], vetv.name()))

        # Очистка выходных параметров ветвей с нулевым сопротивлением.
        self.del_zero_vetv()

        # Очистка выходных параметров ветвей с бесконечным сопротивлением.
        self.del_big_vetv()

        # Очистка выходных параметров ветвей с одинаковым началом и концом.
        for vetv in self.del_circle_vetv():
            warnings.append(u'%s: %s "%s", %s' % (errors.messages['warning'], errors.messages['deleted element'], vetv,
                                                  errors.messages['its start and end nodes are the same']))

        # Создание списка узлов.
        self.create_usli()
        # Удаление ветвей, соединенных с "пустыми" узлами.
        self.del_not_vetv()
        return errs, warnings, messages

    def create_usli(self):
        # Очистка узлов от предыдущих результатов расчета.
        for usel in self.usli.children_list():
            usel = self.usli.find_child(usel)
            usel.output_parameters().remove_children()
        # Проход по ветвям и запись списка узлов.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            output_parameters = vetv.output_parameters()
            if output_parameters.children_list():
                begin = output_parameters.begin().value()
                end = output_parameters.end().value()
                if begin and end:
                    # Если начало ветви не ноль.
                    if begin != '0':
                        # Если узел еще не добавлен, то добавляем его и добавляем ветвь (с минусом) в список его ветвей.
                        if begin not in self.usli.children_list():
                            self.usli.add_child(classes.Usel(node_name=begin, parent=self.usli))
                        usel = self.usli.find_child(begin)
                        usel_output_parameters = usel.output_parameters()
                        # Добавляем ветвь (с минусом) в список ветвей этого узла.
                        vetvi = usel_output_parameters.vetvi().value()
                        vetvi.append(u'-%s' % vetv.name())
                        usel_output_parameters.vetvi().set_value(vetvi)
                    # Если конец ветви не ноль.
                    if end != '0':
                        # Если узел еще не добавлен, то добавляем его и добавляем ветвь (с минусом) в список его ветвей.
                        if end not in self.usli.children_list():
                            self.usli.add_child(classes.Usel(node_name=end, parent=self.usli))
                        usel = self.usli.find_child(end)
                        usel_output_parameters = usel.output_parameters()
                        # Добавляем ветвь (с минусом) в список ветвей этого узла.
                        vetvi = usel_output_parameters.vetvi().value()
                        vetvi.append(vetv.name())
                        usel_output_parameters.vetvi().set_value(vetvi)

    def del_not_vetv(self):
        # Удаление ветвей, соединенных с "пустыми" узлами. Сам узел тоже удаляется.
        # Пустой узел - узел, к которому присоединена только одна ветвь.
        del_list = []
        renew = True
        while renew:
            # Цикл повторяется каждый раз, когда находится ветвь соединенная с "пустыми" узлами
            renew = False
            for vetv in self.vetvi.children_list():
                # Проход по всем ветвям
                vetv = self.vetvi.find_child(vetv)
                output_parameters = vetv.output_parameters()
                if output_parameters.children_list():
                    begin = output_parameters.begin().value()
                    end = output_parameters.end().value()
                    if begin != u'0' and end != u'0':
                        # Если начальный узел и конечный узел не нулевые.
                        begin_usel = self.usli.find_child(begin)
                        end_usel = self.usli.find_child(end)
                        del_begin = False
                        if begin_usel:
                            begin_usel_output_parameters = begin_usel.output_parameters()
                            # Если начальный узел найден.
                            if len(begin_usel_output_parameters.vetvi().value()) <= 1:
                                # Если у начального узла в списке ветвей числится меньше одной ветви,
                                # то выходные параметры ветви и узла удаляются.
                                output_parameters.remove_children()
                                begin_usel_output_parameters.remove_children()
                                # Удаление имени ветви из списка ветвей конечного узла.
                                if end_usel:
                                    end_usel_output_parameters = end_usel.output_parameters()
                                    vetvi = end_usel_output_parameters.vetvi().value()
                                    if vetv.name() in vetvi:
                                        vetvi.remove(vetv.name())
                                    if not vetvi:
                                        # Если к узлу не присоединены другие ветви, то параметры узла удаляются.
                                        end_usel_output_parameters.remove_children()
                                    else:
                                        end_usel_output_parameters.vetvi().set_value(vetvi)
                                del_list.append(vetv.name())
                                renew = True
                                del_begin = True
                        if end_usel and not del_begin:
                            end_usel_output_parameters = end_usel.output_parameters()
                            if len(end_usel_output_parameters.vetvi().value()) <= 1:
                                # Если у конечного узла в списке ветвей числится меньше одной ветви,
                                # то выходные параметры ветви и узла удаляются.
                                output_parameters.remove_children()
                                end_usel_output_parameters.remove_children()
                                # Удаление имени ветви из списка ветвей начального узла.
                                if begin_usel:
                                    begin_usel_output_parameters = begin_usel.output_parameters()
                                    vetvi = begin_usel_output_parameters.vetvi().value()
                                    if u'-%s' % vetv.name() in vetvi:
                                        vetvi.remove(u'-%s' % vetv.name())
                                    if not vetvi:
                                        # Если к узлу не присоединены другие ветви, то параметры узла удаляются.
                                        begin_usel_output_parameters.remove_children()
                                    else:
                                        begin_usel_output_parameters.vetvi().set_value(vetvi)
                                del_list.append(vetv.name())
                                renew = True
        return del_list

    def del_big_vetv(self):
        # Очистка выходных параметров ветвей с бесконечным сопротивлением (проводимость 1E-30).
        del_list = []
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            output_parameters = vetv.output_parameters()
            # Поиск ветвей с бесконечным сопротивлением.
            if output_parameters.children_list():
                y = output_parameters.y().value()
                if y == 1E-30:
                    # Удаление выходных параметров ветви.
                    output_parameters.remove_children()
                    del_list.append(vetv.name())
        return del_list

    def del_zero_vetv(self):
        # Очистка выходных параметров ветвей с нулевым сопротивлением (проводимость 1E30).
        del_list = []
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            output_parameters = vetv.output_parameters()
            # Поиск ветвей с нулевым сопротивлением.
            if output_parameters.children_list():
                y = output_parameters.y().value()
                if y == 1E30:
                    begin = output_parameters.begin().value()
                    end = output_parameters.end().value()
                    # Удаление выходных параметров ветви.
                    output_parameters.remove_children()
                    del_list.append([vetv.name(), end])
                    if begin != '0' and end == '0':
                        end_v = begin
                        begin_v = '0'
                    else:
                        begin_v = begin
                        end_v = end
                    # Проход по ветвям и перемещение их в начальный узел,
                    # если начало и конец удаленной ветви не ноль.
                    if not (begin_v == '0' and end_v == '0'):
                        for vetv_1 in self.vetvi.children_list():
                            vetv_1 = self.vetvi.find_child(vetv_1)
                            output_parameters_1 = vetv_1.output_parameters()
                            if output_parameters_1.children_list():
                                begin_1 = output_parameters_1.begin().value()
                                end_1 = output_parameters_1.end().value()
                                if begin_1 and end_1:
                                    if begin_1 == end_v:
                                        output_parameters_1.begin().set_value(begin_v)
                                    if end_1 == end_v:
                                        output_parameters_1.end().set_value(begin_v)
        return del_list

    def del_circle_vetv(self):
        # Очистка выходных параметров ветвей с одинаковым началом и концом.
        # Если ветвь начинается и заканчивается в нуле, то она не удаляется.
        del_list = []
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            output_parameters = vetv.output_parameters()
            #поиск ветвей с одинаковым началом и концом
            if output_parameters.children_list():
                begin = output_parameters.begin().value()
                end = output_parameters.end().value()
                if begin and end:
                    if begin == end:
                        # Удаление ветви из списка ветвей в узле, если узел не нулевой.
                        if begin != '0':
                            # Удаление выходных параметров ветви.
                            output_parameters.remove_children()
                            del_list.append(vetv.name())
        return del_list

    def create_a_matrix(self, usel_num):
        # Создание квадратной матрицы коэффициентов (a).
        a = zeros((usel_num, usel_num), dtype='complex128')
        # Номер строки матрицы.
        ln = 0
        # Проход по строкам матрицы.
        for l in self.usli.children_list():
            l = self.usli.find_child(l)
            l_output_parameters = l.output_parameters()
            if l_output_parameters.children_list():
                # Номер столбца матрицы.
                usel_l_vetvi = l_output_parameters.vetvi().value()
                kn = 0
                # Проход по столбцам матрицы.
                for k in self.usli.children_list():
                    k = self.usli.find_child(k)
                    k_output_parameters = k.output_parameters()
                    if k_output_parameters.children_list():
                        usel_k_vetvi = k_output_parameters.vetvi().value()
                        # Сумма проводимостей ветвей присоедиенных к узлу.
                        if k.name() == l.name():
                            for m in usel_l_vetvi:
                                if m[0] == '-':
                                    vetv_name = m[1:]
                                else:
                                    vetv_name = m
                                vetv = self.vetvi.find_child(vetv_name)
                                a[ln][kn] += vetv.output_parameters().y().value()
                        # Сумма проводимостей прямых ветвей между узлами k и l.
                        else:
                            v_k = set()
                            for us in usel_k_vetvi:
                                if us[0] == '-':
                                    v_k.add(us[1:])
                                else:
                                    v_k.add(us)
                            v_l = set()
                            for us in usel_l_vetvi:
                                if us[0] == '-':
                                    v_l.add(us[1:])
                                else:
                                    v_l.add(us)
                            for m in list(v_k & v_l):
                                if m[0] == '-':
                                    vetv_name = m[1:]
                                else:
                                    vetv_name = m
                                vetv = self.vetvi.find_child(vetv_name)
                                a[ln][kn] -= vetv.output_parameters().y().value()
                        kn += 1
                ln += 1
        return a

    def create_b_matrix(self, usel_num):
        # Создание матрицы источников (b).
        b = zeros((usel_num, 1), dtype='complex128')
        kn = 0
        # Проход по столбцам матрицы.
        for usel in self.usli.children_list():
            usel = self.usli.find_child(usel)
            usel_output_parameters = usel.output_parameters()
            if usel_output_parameters.children_list():
                usel_vetvi = usel_output_parameters.vetvi().value()
                for vetv_name in usel_vetvi:
                    if vetv_name[0] == '-':
                        # Если источник направлен к узлу, то +.
                        minus = -1
                        vetv_name = vetv_name[1:]
                    else:
                        # Если источник направлен от узла, то -.
                        minus = 1
                    vetv = self.vetvi.find_child(vetv_name)
                    vetv_output_parameters = vetv.output_parameters()
                    if vetv.input_parameters().vetv_type().value() in prog_settings.NAGR_EDS:
                        minus *= -1
                    e = vetv_output_parameters.e().value()
                    y = vetv_output_parameters.y().value()
                    j = vetv_output_parameters.j().value()
                    b[kn][0] += minus * (e * y + j)
                kn += 1
        return b

    def vetv_consumption_regulation(self, vetv=classes.Vetv(''), vetv_output_parameters=classes.OutputParameters(),
                                    vetv_type=''):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Регулировка вевти нагрузки.
        shod = 1
        reg = 0
        # напряжение на ветви.
        u = vetv_output_parameters.u1().value() - vetv_output_parameters.u2().value()
        if abs(u) < 1E-6:
            warnings.append(u'%s: %s "%s". %s' % (errors.messages['warning'],
                                                  errors.messages['can not be achieved a given power'],
                                                  vetv.name(), errors.messages['maybe not enough power sources']))
            self.n_shod = self.settings.max_shod + 1
        else:
            # мощность ветви S = U * sopr(U * Y).
            s = u * (u * vetv_output_parameters.y().value()).conjugate()
            # Номинальная мощность ветви.
            vetv_s = vetv_output_parameters.s().value()
            # если PQ нагрузка не отрегулирована.
            if vetv_type == 'PQ_post':
                if abs(vetv_s - s) > self.settings.delta_norm:
                    s = vetv_s
                    reg = 1
            # если P нагрузка не отрегулирована.
            elif vetv_type == 'P_post':
                if abs(vetv_s.real - s.real) > self.settings.delta_norm:
                    s = vetv_s.real + s.imag * 1j
                    reg = 1
            # если Q нагрузка не отрегулирована.
            elif vetv_type == 'Q_post':
                if abs(vetv_s.imag - s.imag) > self.settings.delta_norm:
                    s = s.real + vetv_s.imag * 1j
                    reg = 1
            # если произведена регулировка, то производится запись нового значения Y.
            if reg == 1:
                # Y = sopr(S / U) / U.
                vetv_output_parameters.y().set_value((s / u).conjugate() / u)
                shod = 0
        return errs, warnings, messages, shod

    def vetv_consumption_eds_regulation(self, vetv=classes.Vetv(''), vetv_output_parameters=classes.OutputParameters(),
                                        vetv_type=''):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Регулировка ветви нагрузки с ЭДС.
        shod = 1
        reg = 0
        # напряжение на ветви.
        u = vetv_output_parameters.u1().value() - vetv_output_parameters.u2().value()
        if abs(u) < 1E-6:
            warnings.append(u'%s: %s "%s". %s' % (errors.messages['warning'],
                                                  errors.messages['can not be achieved a given power'],
                                                  vetv.name(), errors.messages['maybe not enough power sources']))
            self.n_shod = self.settings.max_shod + 1
        else:
            # ток ветви I = (U - E) * Y.
            i = (u - vetv_output_parameters.e().value()) * vetv_output_parameters.y().value()
            # S = U * sopr(I).
            s = u * i.conjugate()
            # Номинальная мощность.
            vetv_s = vetv_output_parameters.s().value()
            if vetv_type == 'nagr_EDS_P_post':
                if abs(vetv_s.real - s.real) > self.settings.delta_norm:
                    s = vetv_s.real + s.imag * 1j
                    reg = 1
            # если произведена регулировка, то производится запись нового значения E.
            if reg == 1:
                # E = U - sopr(S)/(sopr(U)*Y).
                e = u - s.conjugate() / (u.conjugate() * vetv_output_parameters.y().value())
                if abs(e - vetv_output_parameters.e().value()) > 1E-10:
                    vetv_output_parameters.e().set_value(e)
                    shod = 0
                else:
                    warnings.append(u'%s: %s "%s"' % (errors.messages['warning'],
                                                      errors.messages['can not be achieved a given power'],
                                                      vetv.name()))
                    self.n_shod = self.settings.max_shod + 1
        return errs, warnings, messages, shod

    def vetv_generator_regulation(self, vetv=classes.Vetv(''), vetv_output_parameters=classes.OutputParameters(),
                                  vetv_type=''):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Регулирование ветви источника.
        shod = 1
        reg = 0
        # напряжение на ветви.
        u = vetv_output_parameters.u2().value() - vetv_output_parameters.u1().value()
        # ток ветви I = (E - U) * Y.
        i = (vetv_output_parameters.e().value() - u) * vetv_output_parameters.y().value()
        # S = U * sopr(I).
        s = u * i.conjugate()
        # Номинальное напряжение.
        vetv_u = vetv_output_parameters.u().value()
        # Номинальная мощность.
        vetv_s = vetv_output_parameters.s().value()
        # если источник ЭДС (U) не отрегулирован.
        if vetv_type == 'EDS_U_post':
            if abs(vetv_u - u) > self.settings.delta_norm:
                reg = 1
        # если источник ЭДС (PU) не отрегулирован.
        elif vetv_type == 'EDS_PU_post':
            if abs(s.real - vetv_s.real) > self.settings.delta_norm and abs(u - vetv_u) > self.settings.delta_norm:
                s = vetv_s.real + s.imag * 1j
                reg = 1
        # если источник ЭДС (QU) не отрегулирован.
        elif vetv_type == 'EDS_QU_post':
            if abs(s.imag - vetv_s.imag) > self.settings.delta_norm and abs(u - vetv_u) > self.settings.delta_norm:
                s = s.real + vetv_s.imag * 1j
                reg = 1
        # если источник ЭДС (PQ) не отрегулирован.
        elif vetv_type == 'EDS_PQ_post' and abs(s - vetv_s) > self.settings.delta_norm:
            s = vetv_s
            vetv_u = u
            reg = 1
        # если произведена регулировка, то производится запись нового значения E.
        if reg == 1:
            # E = Uн + sopr(S / U) / Y.
            e = vetv_u + (s / u).conjugate() / vetv_output_parameters.y().value()
            if abs(e - vetv_output_parameters.e().value()) > 1E-10:
                vetv_output_parameters.e().set_value(e)
                shod = 0
            else:
                warnings.append(u'%s: %s "%s"' % (errors.messages['warning'],
                                                  errors.messages['can not adjust the source'], vetv.name()))
                self.n_shod = self.settings.max_shod + 1
        return errs, warnings, messages, shod

    def vetv_generator_j_regulation(self, vetv=classes.Vetv(''), vetv_output_parameters=classes.OutputParameters(),
                                    vetv_type=''):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        shod = 1
        reg = 0
        # напряжение на ветви.
        u = vetv_output_parameters.u2().value() - vetv_output_parameters.u1().value()
        # ток ветви J.
        j = vetv_output_parameters.j().value()
        s = u * j.conjugate()
        # Номинальное напряжение.
        vetv_u = vetv_output_parameters.u().value()
        # Номинальная мощность.
        vetv_s = vetv_output_parameters.s().value()
        #если источник ЭДС (U) не отрегулирован
        if vetv_type == 'ist_toka_U_post':
            if abs(vetv_u - u) > self.settings.delta_norm:
                reg = 1
        #если произведена регулировка, то производится запись нового значения J
        if reg == 1:
            # J = U * J / Uн
            j = vetv_u.conjugate() * j / u.conjugate()
            if abs(j - vetv_output_parameters.j().value()) > 1E-10:
                vetv_output_parameters.j().set_value(j)
                shod = 0
            else:
                warnings.append(u'%s: %s "%s"' % (errors.messages['warning'],
                                                  errors.messages['can not adjust the source'], vetv.name()))
                self.n_shod = self.settings.max_shod + 1
        return errs, warnings, messages, shod

    def vetv_rx_voltage_addition_regulation(self, vetv=classes.Vetv(''),
                                            vetv_output_parameters=classes.OutputParameters(), vetv_type=''):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Регулирование ветви источника.
        shod = 1
        reg = 0
        # напряжение на ветви.
        u = vetv_output_parameters.u1().value() - vetv_output_parameters.u2().value()
        # ток ветви I = (U + E) * Y.
        i = (u + vetv_output_parameters.e().value()) * vetv_output_parameters.y().value()
        # Напряжение в середине ветви.
        u1_ = vetv_output_parameters.u1().value() - i / vetv_output_parameters.y().value()
        # Модуль добавки напряжения.
        vetv_input_parameters = vetv.input_parameters()
        vetv_u_addition_module = vetv_input_parameters.voltage_addition().value()
        # Фаза добавки напряжения.
        vetv_u_addition_phase = vetv_input_parameters.voltage_addition_psi().value()
        # если источник ЭДС (U) не отрегулирован.
        u2_module = vetv_u_addition_module * abs(u1_) / 100. + abs(u1_)
        u2_phase = math.radians(vetv_u_addition_phase) + cmath.phase(u1_)
        u2_ = cmath.rect(u2_module, u2_phase)
        if vetv_type == 'RX_voltage_addition':
            if abs(vetv_output_parameters.u2().value() - u2_) > self.settings.delta_norm:
                reg = 1
        # если произведена регулировка, то производится запись нового значения E.
        if reg == 1:
            # E = U2' - U1 + I / Y.
            e = u2_ - vetv_output_parameters.u1().value() + i / vetv_output_parameters.y().value()
            if abs(e - vetv_output_parameters.e().value()) > 1E-10:
                vetv_output_parameters.e().set_value(e)
                shod = 0
            else:
                warnings.append(u'%s: %s "%s"' % (errors.messages['warning'],
                                                  errors.messages['can not adjust the source'], vetv.name()))
                self.n_shod = self.settings.max_shod + 1
        return errs, warnings, messages, shod

    def calculate(self, progress_bar=None, progress_bar_start=0, progress_bar_end=100):
        # Расчет цепи.
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Количество узлов.
        usel_num = 0
        for usel in self.usli.children_list():
            usel = self.usli.find_child(usel)
            if usel.output_parameters().children_list():
                usel_num += 1
        if progress_bar is not None:
            progress_bar_length = (progress_bar_end - progress_bar_start) / 100.
            progress_bar.setValue(5 * progress_bar_length + progress_bar_start)

        shod = 0
        self.n_shod = 0
        while shod == 0 and self.n_shod < self.settings.max_shod:
            shod = 1
            self.n_shod += 1
            #1 Запись и расчет матриц системы уравнений.
            #1.1 Создание квадратной матрицы коэффициентов (a).
            a = self.create_a_matrix(usel_num)
            #1.2 Создание матрицы источников (b).
            b = self.create_b_matrix(usel_num)
            if progress_bar is not None:
                progress_bar.setValue(10 * progress_bar_length + progress_bar_start)
            #2 Рассчет напряжений.
            usli_voltage = linalg.solve(a, b)
            i = 0
            for usel in self.usli.children_list():
                usel = self.usli.find_child(usel)
                usel_output_parameters = usel.output_parameters()
                if usel_output_parameters.children_list():
                    usel_output_parameters.u().set_value(usli_voltage[i][0])
                    i += 1
            if progress_bar is not None:
                progress_bar.setValue(20 * progress_bar_length + progress_bar_start)
            #3 Запись напряжений в началах и концах ветвей и пересчет параметров переменных ветвей
            # при отклонении от заданных значений.
            for vetv in self.vetvi.children_list():
                vetv = self.vetvi.find_child(vetv)
                vetv_output_parameters = vetv.output_parameters()
                if vetv_output_parameters.children_list():
                    #3.1 Запись напряжения в конце и начале ветви
                    if vetv_output_parameters.begin().value() != '0':
                        usel = self.usli.find_child(vetv_output_parameters.begin().value())
                        vetv_output_parameters.u1().set_value(usel.output_parameters().u().value())
                    if vetv_output_parameters.end().value() != '0':
                        usel = self.usli.find_child(vetv_output_parameters.end().value())
                        vetv_output_parameters.u2().set_value(usel.output_parameters().u().value())
                    if progress_bar is not None:
                        progress_bar_start_2 = 20 * progress_bar_length + progress_bar_start
                        progress_bar_end_2 = 100 * progress_bar_length + progress_bar_start
                        progress_bar_length_2 = (progress_bar_end_2 - progress_bar_start_2) / 100.
                        progress_bar.setValue(50 * progress_bar_length_2 + progress_bar_start_2)
                    #3.2 Проверка по типам ветвей.
                    # Регулировка параметров ветвей.
                    vetv_type = vetv.input_parameters().vetv_type().value()
                    #- ветвь с нагрузкой с заданными значениями PQ или P или Q (изменяется Y).
                    if vetv_type in ('PQ_post', 'P_post', 'Q_post'):
                        result = self.vetv_consumption_regulation(vetv, vetv_output_parameters, vetv_type)
                        if not result[3]:
                            shod = 0
                        errs += result[0]
                        warnings += result[1]
                        messages += result[2]
                    #- ветвь с нагрузкой с ЭДС заданным значением P (изменяется E).
                    elif vetv_type in ('nagr_EDS', 'nagr_EDS_P_post', 'nagr_EDS_Q_post', 'nagr_EDS_PQ_post'):
                        result = self.vetv_consumption_eds_regulation(vetv, vetv_output_parameters, vetv_type)
                        if not result[3]:
                            shod = 0
                        errs += result[0]
                        warnings += result[1]
                        messages += result[2]
                    #- ветвь с источником ЭДС с заданым значением U или PU или QU или PQ (изменяется E)
                    elif vetv_type in ('EDS_U_post', 'EDS_PU_post', 'EDS_QU_post', 'EDS_PQ_post'):
                        result = self.vetv_generator_regulation(vetv, vetv_output_parameters, vetv_type)
                        if not result[3]:
                            shod = 0
                        errs += result[0]
                        warnings += result[1]
                        messages += result[2]
                    #- ветвь с источником тока с заданым значением U или PU или QU или PQ (изменяется I)
                    elif vetv_type == 'ist_toka_U_post':
                        result = self.vetv_generator_j_regulation(vetv, vetv_output_parameters, vetv_type)
                        if not result[3]:
                            shod = 0
                        errs += result[0]
                        warnings += result[1]
                        messages += result[2]
                    elif vetv_type == 'RX_voltage_addition':
                        result = self.vetv_rx_voltage_addition_regulation(vetv, vetv_output_parameters, vetv_type)
                        if not result[3]:
                            shod = 0
                        errs += result[0]
                        warnings += result[1]
                        messages += result[2]

                    if progress_bar is not None:
                        progress_bar.setValue(100 * progress_bar_length_2 + progress_bar_start_2)

        if self.n_shod == self.settings.max_shod:
            warnings.append(u'%s: %s' % (language.Errors().messages['maximum number of iterations'],
                                         self.settings.max_shod))

        if progress_bar is not None:
            progress_bar.setValue(100 * progress_bar_length + progress_bar_start)
        messages.append(u'%s' % language.Errors().messages['complete main calculations'])
        return errs, warnings, messages

    def calc_vetvi_voltage_real(self):
        # Расчет реальных напряжений в началах и концах ветвей.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            vetv_output_parameters = vetv.output_parameters()
            if vetv_output_parameters.children_list():
                u1 = vetv_output_parameters.u1().value()
                u2 = vetv_output_parameters.u2().value()
                if vetv.input_parameters().vetv_type().value() in (prog_settings.EDS + prog_settings.II):
                    u = u2 - u1
                else:
                    u = u1 - u2
                k_p = vetv_output_parameters.k_p().value()
                vetv_output_parameters.real_u1().set_value(u1 / k_p)
                vetv_output_parameters.real_u2().set_value(u2 / k_p)
                vetv_output_parameters.real_u().set_value(u / k_p)
        return True

    def calc_eds_real(self):
        # Расчет реальных ЭДС.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            vetv_output_parameters = vetv.output_parameters()
            if vetv_output_parameters.children_list():
                e = vetv_output_parameters.e().value()
                k_p = vetv_output_parameters.k_p().value()
                vetv_output_parameters.real_e().set_value(e / k_p)
        return True

    def calc_vetvi_current(self):
        # Расчет приведенных токов.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            vetv_output_parameters = vetv.output_parameters()
            if vetv_output_parameters.children_list():
                vetv_type = vetv.input_parameters().vetv_type().value()
                j = vetv_output_parameters.j().value()
                u1 = vetv_output_parameters.u1().value()
                u2 = vetv_output_parameters.u2().value()
                y = vetv_output_parameters.y().value()
                e = vetv_output_parameters.e().value()
                # Если источник тока.
                if vetv_type in prog_settings.II:
                    vetv_output_parameters.i().set_value(j)
                #если постоянное сопротивление или нагрузка I = U12 * Y
                elif vetv_type in prog_settings.NAGR or vetv_type == 'RX_post':
                    vetv_output_parameters.i().set_value((u1 - u2) * y)
                #если источник ЭДС I = (E - U21) * Y
                elif vetv_type in (prog_settings.EDS + ['RX_voltage_addition']):
                    vetv_output_parameters.i().set_value((e - (u2 - u1)) * y)
                #если нагрузка с ЭДС I = (-E + U12) * Y
                elif vetv_type in prog_settings.NAGR_EDS:
                    vetv_output_parameters.i().set_value((-e + (u1 - u2)) * y)
        return True

    def calc_vetvi_current_real(self):
        # Расчет реальных токов.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            vetv_output_parameters = vetv.output_parameters()
            if vetv_output_parameters.children_list():
                i = vetv_output_parameters.i().value()
                k_p = vetv_output_parameters.k_p().value()
                vetv_output_parameters.real_i().set_value(i * k_p.conjugate())
        return True

    def calc_power(self):
        # Расчет мощностей приемников и источников.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            vetv_output_parameters = vetv.output_parameters()
            if vetv_output_parameters.children_list():
                vetv_type = vetv.input_parameters().vetv_type().value()
                i = vetv_output_parameters.i().value()
                u1 = vetv_output_parameters.u1().value()
                u2 = vetv_output_parameters.u2().value()
                y = vetv_output_parameters.y().value()
                e = vetv_output_parameters.e().value()
                #1 Расчет мощностей приемников
                #если нагрузка или ветвь с постоянным сопротивлением S = sopr(I) * U12
                if vetv_type in prog_settings.NAGR or vetv_type == 'RX_post':
                    vetv_output_parameters.priem_s().set_value(i.conjugate() * (u1 - u2))
                #если ветвь с источником ЭДС S = I * sopr(I) / Y
                elif vetv_type in (prog_settings.EDS + ['RX_voltage_addition']):
                    if y:
                        vetv_output_parameters.priem_s().set_value(i * i.conjugate() / y)
                    else:
                        vetv_output_parameters.priem_s().set_value(0. + 0.j)
                #если ветвь с нагрузкой с источником ЭДС S = sopr(I) * U12
                elif vetv_type in prog_settings.NAGR_EDS:
                    vetv_output_parameters.priem_s().set_value(i.conjugate() * (u1 - u2))
                else:
                    vetv_output_parameters.priem_s().set_value(0. + 0.j)

                #2 Расчет мощностей источников
                #если источник ЭДС S = E * sopr(I)
                if vetv_type in (prog_settings.EDS + ['RX_voltage_addition']):
                    vetv_output_parameters.ist_s().set_value(e * i.conjugate())
                #если источник тока S = sopr(I) * U21
                elif vetv_type in prog_settings.II:
                    vetv_output_parameters.ist_s().set_value(i.conjugate() * (u2 - u1))
                else:
                    vetv_output_parameters.ist_s().set_value(0. + 0.j)
        return True

    def calc_usli_voltage_real(self):
        # Расчет реальных напряжений в узлах.
        for usel in self.usli.children_list():
            usel = self.usli.find_child(usel)
            usel_output_parameters = usel.output_parameters()
            if usel_output_parameters.children_list():
                for vetv in usel_output_parameters.vetvi().value():
                    if vetv[0] == '-':
                        vetv_name = vetv[1:]
                    else:
                        vetv_name = vetv
                    vetv = self.vetvi.find_child(vetv_name)
                    vetv_output_parameters = vetv.output_parameters()
                    if vetv_output_parameters.begin().value() == usel.name():
                        u = vetv_output_parameters.u1().value() / vetv_output_parameters.k_p().value()
                    else:
                        u = vetv_output_parameters.u2().value() / vetv_output_parameters.k_p().value()
                    if u not in usel_output_parameters.real_u_usel().value():
                        u_1 = usel_output_parameters.real_u_usel().value()
                        u_1.append(u)
                        usel_output_parameters.real_u_usel().set_value(u_1)
        return True

    def calc_vetvi_power(self):
        # Расчет мощностей в началах и концах ветвей.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            vetv_output_parameters = vetv.output_parameters()
            if vetv_output_parameters.children_list():
                i = vetv_output_parameters.i().value()
                u1 = vetv_output_parameters.u1().value()
                u2 = vetv_output_parameters.u2().value()
                vetv_output_parameters.s1().set_value(i.conjugate() * u1)
                vetv_output_parameters.s2().set_value(i.conjugate() * u2)
        return True

    def calc_add(self):
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Реальные напряжения на ветвях.
        self.calc_vetvi_voltage_real()
        # Ральные ЭДС.
        self.calc_eds_real()
        # Приведенные токи в ветвях.
        self.calc_vetvi_current()
        # Реальные токи в ветвях.
        self.calc_vetvi_current_real()
        # Мощности приемников и источников в ветвях.
        self.calc_power()
        # Реальные напряжения в узлах.
        self.calc_usli_voltage_real()
        # Потоки мощностей в ветвях.
        self.calc_vetvi_power()
        self.correct_units()
        messages.append(u'%s' % language.Errors().messages['complete additional calculations'])
        return errs, warnings, messages

    def results_check(self):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Проверка по балансу мощностей.
        sum_ist_power = 0
        sum_priem_power = 0
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            vetv_output_parameters = vetv.output_parameters()
            if vetv_output_parameters.children_list():
                sum_ist_power += vetv_output_parameters.ist_s().value()
                sum_priem_power += vetv_output_parameters.priem_s().value()
        if sum_ist_power and sum_priem_power:
            delta_s = sum_ist_power - sum_priem_power
            delta_s = round(delta_s.real, 5) + round(delta_s.imag, 5)*1j
            if abs(delta_s) < 1E-3:
                messages.append(u'%s %s' % (errors.messages['pass check balance power'], delta_s))
            else:
                warnings.append(u'%s %s' % (errors.messages['fail check balance power'], delta_s))

        # Проверка по 1 закону Кирхгофа.
        kirh = True
        for usel in self.usli.children_list():
            usel = self.usli.find_child(usel)
            usel_output_parameters = usel.output_parameters()
            if usel_output_parameters.children_list():
                delta_i = 0
                for vetv in usel_output_parameters.vetvi().value():
                    if vetv[0] == '-':
                        minus = -1
                        vetv_name = vetv[1:]
                    else:
                        minus = 1
                        vetv_name = vetv
                    vetv = self.vetvi.find_child(vetv_name)
                    delta_i += minus * vetv.output_parameters().i().value()
                delta_i = round(delta_i.real, 5) + round(delta_i.imag, 5)*1j
                if abs(delta_i) > 1E-3:
                    kirh = False
                    warnings.append(u'%s "%s": %s' % (errors.messages['fail check 1 Kirh'], usel.name(), delta_i))
        if kirh:
            messages.append(u'%s' % errors.messages['pass check 1 Kirh'])
        return errs, warnings, messages

    def correct_units(self):
        # Обратное преобразование единиц измерения.
        voltage = ['e', 'real_u', 'real_e', 'u', 'u1', 'real_u1', 'u2', 'real_u2', 'real_voltage']
        current = ['i', 'real_i', 'j']
        power = ['s', 's1', 's2', 'ist_s', 'priem_s']
        self.base_u /= self.unit_k_voltage
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            vetv_output_parameters = vetv.output_parameters()
            for param in vetv_output_parameters.children_list():
                param = vetv_output_parameters.find_child(param)
                param_value = param.value()
                if param.name() in voltage:
                    param.set_value(param_value / self.unit_k_voltage)
                elif param.name() in current:
                    param.set_value(param_value / self.unit_k_current)
                elif param.name() in power:
                    param.set_value(param_value / self.unit_k_full_power)

        for usel in self.usli.children_list():
            usel = self.usli.find_child(usel)
            usel_output_parameters = usel.output_parameters()
            for param in usel_output_parameters.children_list():
                param = usel_output_parameters.find_child(param)
                param_value = param.value()
                if param.name() == 'u':
                    param.set_value(param_value / self.unit_k_voltage)
                if param.name() == 'real_u_usel':
                    for i in range(len(param_value)):
                        param_value[i] /= self.unit_k_voltage
                    param.set_value(param_value)
        return True
