# -*- coding: utf-8 -*-
""" Модуль дополняет новыми методами для расширенных расчетов класс, описанный в модуле node_volt_calc,
Модуль также описывает функции отвечающие за вывод результатов расчета на экран или в файл.
"""

import cmath
import cPickle
import csv
import math
from xml.dom.minidom import Document
import xml.etree.ElementTree
import xml.etree.cElementTree as ET
import classes
import language
import node_volt_calc as nv
import prog_settings

SCHEME_TYPES = ['', '1_phase', '3_phase', '3_phase_a']


class NodeScheme(nv.NodeScheme):
    
    def __init__(self, settings=nv.Settings()):
        nv.NodeScheme.__init__(self, settings)
        self.base_u_input = '1.0'
        self.scheme_file_name = ''  # имя файла схемы
        self.scheme_type = '3_phase'    # тип схемы
        
        self.unit_active_power = u'Вт'  # единица измерения активной мощности
        self.unit_reactive_power = u'вар'   # единица измерения реактивной мощности
        self.unit_full_power = u'ВА'    # единица измерения полной мощности
        self.unit_voltage = u'В'    # единица измерения напряжения
        self.unit_current = u'А'    # единица измерения тока
        
        self.print_power = True
        self.print_vetvi_power = True
        self.print_vetvi_current = True
        self.print_vetvi_voltage = True
        self.print_usli_voltage = True

        self.use_temperature_coefficient = False    # учитывать или нет температурный коэффициент сопротивления

    def read_scheme_file_nvs(self):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        if not self.scheme_file_name:
            errs.append(u'%s: %s' % (errors.messages['error'], errors.messages['not set scheme file']))
            return errs, warnings, messages

        # Попытка открытия файла
        try:
            open_params = cPickle.load(open(self.scheme_file_name, 'rb'))
        except:
            errs.append(u'%s: %s "%s"' % (errors.messages['error'], errors.messages['unable to open scheme file'],
                                          self.scheme_file_name))
            return errs, warnings, messages

        self.vetvi = classes.Parameters('vetvi')
        self.usli = classes.Parameters('usli')
        self.base_u_input = '1.0'
        self.scheme_type = ''

        def read_param(param_name, def_param, err=u''):
            try:
                return open_params[param_name]
            except:
                if err:
                    warnings.append(err)
                return def_param
        # Считывание параметров из файла.
        self.base_u_input = read_param('baseU', '1.0',
                                       u'%s: %s' % (errors.messages['warning'],
                                                    errors.messages['can not read the base voltage']))
        self.unit_k_active_power = read_param('unit_k_active_power', 1.)
        self.unit_k_reactive_power = read_param('unit_k_reactive_power', 1.)
        self.unit_k_full_power = read_param('unit_k_full_power', 1.)
        self.unit_k_voltage = read_param('unit_k_voltage', 1.)
        self.unit_k_current = read_param('unit_k_current', 1.)
        self.print_power = read_param('print_power', True)
        self.print_vetvi_power = read_param('print_vetvi_power', True)
        self.print_vetvi_current = read_param('print_vetvi_current', True)
        self.print_vetvi_voltage = read_param('print_vetvi_voltage', True)
        self.print_usli_voltage = read_param('print_usli_voltage', True)
        self.use_temperature_coefficient = read_param('use_temperature_coefficient', False)
        self.scheme_type = read_param('scheme_type', '')

        self.vetvi = read_param('vetvi', classes.Parameters('vetvi'),
                                u'%s: %s' % (errors.messages['error'],
                                             errors.messages['can not read elements parameters']))
        self.usli = read_param('usli', classes.Parameters('usli'),
                               u'%s: %s' % (errors.messages['error'],
                                            errors.messages['can not read nodes parameters']))

        self.read_units()

        return errs, warnings, messages

    def read_scheme_file_xml(self):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Функция чтения XML-файла.
        if not self.scheme_file_name:
            errs.append(u'%s: %s' % (errors.messages['error'], errors.messages['not set scheme file']))
            return errs, warnings, messages
        # Попытка открытия файла
        try:
            tree = ET.parse(self.scheme_file_name)
        except:
            errs.append(u'%s: %s "%s"' % (errors.messages['error'], errors.messages['unable to open scheme file'],
                                          self.scheme_file_name))
            return errs, warnings, messages

        self.vetvi = classes.Parameters('vetvi')
        self.usli = classes.Parameters('usli')
        self.base_u_input = '1.0'
        self.scheme_type = ''

        root = tree.getroot()

        # Чтение базисного напряжения
        base_u_1 = root.find('baseU')
        if base_u_1 is not None:
            try:
                self.base_u_input = base_u_1.text.replace(',', '.')
            except:
                warnings.append(u'%s: %s' % (errors.messages['warning'],
                                             errors.messages['can not read the base voltage']))

        def read_settings_text(setting, err_text, list_settings, default_setting):
            # Чтение текстовых настроек.
            setting_xml = root.find(setting)
            if setting_xml is not None:
                setting_text = setting_xml.text
                if setting_text is None:
                    setting_text = ''
                if setting_text in list_settings:
                    setting_out = setting_text
                else:
                    warnings.append(err_text)
                    setting_out = default_setting
            else:
                setting_out = default_setting
            return setting_out

        def read_settings_float(setting, err_text, list_settings, default_setting):
            # Чтение текстовых настроек.
            setting_xml = root.find(setting)
            if setting_xml is not None:
                setting_text = setting_xml.text
                if setting_text is None:
                    setting_text = ''
                if setting_text in list_settings:
                    try:
                        setting_out = float(setting_text)
                    except ValueError:
                        warnings.append(err_text)
                        setting_out = default_setting
                else:
                    warnings.append(err_text)
                    setting_out = default_setting
            else:
                setting_out = default_setting
            return setting_out

        units_k = ['1', '1.', '1.0', '1000', '1000.', '1000.0', '1E3', '1000000', '1000000.', '1000000.0', '1E6']

        # Чтение типа схемы.
        scheme_types = [scheme_type[1] for scheme_type in language.Properties().scheme_types]
        self.scheme_type = read_settings_text('scheme_type',
                                              u'%s: %s' % (errors.messages['warning'],
                                                           errors.messages['invalid scheme type']), scheme_types, '')

        # Чтение единицы измерения активной мощности
        self.unit_k_active_power = read_settings_float('unit_k_active_power',
                                                       u'%s: %s' % (errors.messages['warning'],
                                                                    errors.messages['invalid unit of active power']),
                                                       units_k, 1.)

        # Чтение единицы измерения реактивной мощности
        self.unit_k_reactive_power = read_settings_float('unit_k_reactive_power',
                                                         u'%s: %s' % (errors.messages['warning'],
                                                                      errors.messages['invalid unit of reactive '
                                                                                      'power']), units_k, 1.)

        # Чтение единицы измерения полной мощности
        self.unit_k_full_power = read_settings_float('unit_k_full_power',
                                                     u'%s: %s' % (errors.messages['warning'],
                                                                  errors.messages['invalid unit of reactive power']),
                                                     units_k, 1.)

        # Чтение единицы измерения напряжения и ЭДС
        self.unit_k_voltage = read_settings_float('unit_k_voltage',
                                                  u'%s: %s' % (errors.messages['warning'],
                                                               errors.messages['invalid unit of voltage']), units_k, 1.)

        # Чтение единицы измерения тока
        self.unit_k_current = read_settings_float('unit_k_current',
                                                  u'%s: %s' % (errors.messages['warning'],
                                                               errors.messages['invalid unit of current']), units_k, 1.)
        self.read_units()

        # Чтение настроек вывода результатов расчета
        def read_print_settings(setting):
            setting_xml = root.find(setting)
            if setting_xml is not None:
                if setting_xml.text == '1':
                    return True
                elif setting_xml.text == '0':
                    return False
            else:
                return True

        self.print_power = read_print_settings('print_power')
        self.print_vetvi_power = read_print_settings('print_vetvi_power')
        self.print_vetvi_current = read_print_settings('print_vetvi_current')
        self.print_vetvi_voltage = read_print_settings('print_vetvi_voltage')
        self.print_usli_voltage = read_print_settings('print_usli_voltage')

        # Чтение других настроек.
        self.use_temperature_coefficient = read_print_settings('use_temperature_coefficient')

        def read_param(param_xml):
            attribs = param_xml.attrib
            param_name = ''
            param_type = ''
            param_value = ''

            if 'name' in attribs:
                param_name = param_xml.attrib['name']
            if 'type' in attribs:
                param_type = param_xml.attrib['type']
            if 'value' in attribs:
                param_value = param_xml.attrib['value']

            return param_name, param_type, param_value

        def read_group_params(element, element_xml):
            for group_parameters_xml in element_xml:
                if group_parameters_xml.tag == 'input_parameters':
                    group_parameters = element.input_parameters()
                elif group_parameters_xml.tag == 'output_parameters':
                    group_parameters = element.output_parameters()
                elif group_parameters_xml.tag == 'gui_parameters':
                    group_parameters = element.gui_parameters()
                elif group_parameters_xml.tag == 'buffer_parameters':
                    group_parameters = element.buffer_parameters()
                elif group_parameters_xml.tag == 'additional_parameters':
                    group_parameters = element.additional_parameters()
                elif group_parameters_xml.tag == 'optimization_parameters':
                    group_parameters = element.optimization_parameters()
                else:
                    group_parameters = None
                if group_parameters is not None:
                    # Поиск параметров в группе.
                    for parameter in group_parameters_xml:
                        parameter_params = read_param(parameter)
                        param = classes.Parameters(parameter.tag, parameter_params[1], parameter_params[2])
                        group_parameters.add_child(param)

        # Поиск всех элементов в файле.
        vetvi_xml = root.findall('vetv')
        if vetvi_xml is not None:
            for vetv_xml in vetvi_xml:
                # Чтение аттрибутов элементов.
                vetv_params = read_param(vetv_xml)
                if vetv_params[0]:
                    # Добавление нового элемента в схему.
                    vetv = classes.Vetv(vetv_params[0])
                    self.vetvi.add_child(vetv)
                    # Поиск групп параметров элементов.
                    read_group_params(vetv, vetv_xml)
        # Поиск всех узлов в файле.
        usli_xml = root.findall('usel')
        if usli_xml is not None:
            for usel_xml in usli_xml:
                # Чтение аттрибутов узла.
                usel_params = read_param(usel_xml)
                if usel_params[0]:
                    # Добавление нового узла в схему.
                    usel = classes.Usel(usel_params[0])
                    self.usli.add_child(usel)
                    # Поиск групп параметров элемента.
                    read_group_params(usel, usel_xml)

        return errs, warnings, messages

    def write_scheme_file_xml(self):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Функция сохранения XML-файла схемы
        if not self.scheme_file_name:
            errs.append(u'%s: %s' % (errors.messages['error'], errors.messages['not set scheme file']))
            return errs, warnings, messages

        # Создание основной записи <node_volt></node_volt>
        doc = Document()
        doc.encoding = 'UTF-8'
        node_volt_xml = doc.createElement('node_volt')
        doc.appendChild(node_volt_xml)

        # Запись типа схемы <scheme_type></scheme_type>
        scheme_type = doc.createElement('scheme_type')
        node_volt_xml.appendChild(scheme_type)
        scheme_type.appendChild(doc.createTextNode(unicode(self.scheme_type)))

        # Запись базисного напряжения <baseU></baseU>
        base_u = doc.createElement('baseU')
        node_volt_xml.appendChild(base_u)
        base_u.appendChild(doc.createTextNode(unicode(self.base_u_input)))

        # Запись единиц измерения
        def write_unit_settings(setting, value):
            setting_xml = doc.createElement(setting)
            node_volt_xml.appendChild(setting_xml)
            setting_xml.appendChild(doc.createTextNode(unicode(value)))

        write_unit_settings('unit_k_active_power', self.unit_k_active_power)
        write_unit_settings('unit_k_reactive_power', self.unit_k_reactive_power)
        write_unit_settings('unit_k_full_power', self.unit_k_full_power)
        write_unit_settings('unit_k_voltage', self.unit_k_voltage)
        write_unit_settings('unit_k_current', self.unit_k_current)

        # Запись настроек вывода результатов расчета
        def write_print_settings(setting, value):
            setting_xml = doc.createElement(setting)
            node_volt_xml.appendChild(setting_xml)
            if value:
                value = '1'
            else:
                value = '0'
            setting_xml.appendChild(doc.createTextNode(value))

        write_print_settings('print_power', self.print_power)
        write_print_settings('print_vetvi_power', self.print_vetvi_power)
        write_print_settings('print_vetvi_current', self.print_vetvi_current)
        write_print_settings('print_vetvi_voltage', self.print_vetvi_voltage)
        write_print_settings('print_usli_voltage', self.print_usli_voltage)

        write_print_settings('use_temperature_coefficient', self.use_temperature_coefficient)

        def xml_append(parent, element):
            # Функция записи отдельного элемента.
            if element.type() in ('usel', 'vetv'):
                xml_child = doc.createElement(element.type())
                xml_child.setAttribute('name', element.name())
            else:
                xml_child = doc.createElement(element.name())
            parent.appendChild(xml_child)
            xml_child.setAttribute('type', element.type())
            xml_child.setAttribute('value', unicode(element.value()))
            return xml_child

        def xml_node_tree(node):
            # Функция записи дерева элементов.
            for element in node.children_list():
                element = node.find_child(element)
                xml_element = xml_append(node_volt_xml, element)
                if len(element.children_list()):
                    # Запись групп параметров.
                    for parameter_group in element.children_list():
                        parameter_group = element.find_child(parameter_group)
                        xml_parameter_group = xml_append(xml_element, parameter_group)
                        if len(parameter_group.children_list()):
                            # Запись конкретных параметров.
                            for parameter in parameter_group.children_list():
                                parameter = parameter_group.find_child(parameter)
                                xml_append(xml_parameter_group, parameter)
        # Запись параметров элементов и узлов.
        xml_node_tree(self.vetvi)
        xml_node_tree(self.usli)

        try:
            doc.writexml(codecs.open(self.scheme_file_name, 'w', 'utf-8'), indent='', addindent='\t', newl='\n',
                         encoding='utf-8')
            doc.unlink()
        except:
            errs.append(u'%s: %s' % (errors.messages['error'], errors.messages['failed to save scheme file']))

        return errs, warnings, messages

    def write_scheme_file_nvs(self):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        if not self.scheme_file_name:
            errs.append(u'%s: %s' % (errors.messages['error'], errors.messages['not set scheme file']))
        # Запись параметров в файл.
        save_params = {'baseU': self.base_u_input, 'unit_k_active_power': self.unit_k_active_power,
                       'unit_k_reactive_power': self.unit_k_reactive_power, 'unit_k_full_power': self.unit_k_full_power,
                       'unit_k_voltage': self.unit_k_voltage, 'unit_k_current': self.unit_k_current,
                       'print_power': self.print_power, 'print_vetvi_power': self.print_vetvi_power,
                       'print_vetvi_current': self.print_vetvi_current, 'print_vetvi_voltage': self.print_vetvi_voltage,
                       'print_usli_voltage': self.print_usli_voltage, 'vetvi': self.vetvi, 'usli': self.usli,
                       'scheme_type': self.scheme_type, 'use_temperature_coefficient': self.use_temperature_coefficient}
        
        try:
            cPickle.dump(save_params, open(self.scheme_file_name, 'wb'))
        except:
            errs.append(u'%s: %s' % (errors.messages['error'], errors.messages['failed to save scheme file']))

        return errs, warnings, messages

    def insert_element(self, element=None, element_name='', position=0):
        if not element_name:
            element_name = language.Other().vetv_lang
        # Вставка нового элемента.
        if not position:
            # Если не указана позиция, то элемент вставляется в конец.
            position = self.vetvi.child_count()
        for j in range(1, 100):
            element_name_1 = u'%s %s' % (element_name, j)
            if element_name_1 not in self.vetvi.children_list():
                if not isinstance(element, classes.Vetv):
                    element = classes.Vetv(node_name=element_name_1)
                self.vetvi.insert_child(element, position)
                break

    def insert_usel(self, usel=None, usel_name='', position=0):
        if not usel_name:
            usel_name = language.Other().usel_lang
        # Вставка нового узла.
        if not position:
            # Если не указана позиция, то узел вставляется в конец.
            position = self.usli.child_count()
        for j in range(1, 100):
            usel_name_1 = u'%s %s' % (usel_name, j)
            if usel_name_1 not in self.usli.children_list():
                if not isinstance(usel, classes.Usel):
                    usel = classes.Usel(node_name=usel_name_1)
                self.usli.insert_child(usel, position)
                break

    def read_units(self, reverse=None):
        errors = language.Errors()
        properties = language.Properties()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []

        def read_unit(unit, unit_list, def_unit):
            if unit in unit_list:
                return unit_list[unit]
            else:
                warnings.append(u'%s: %s "%s"' % (errors.messages['warning'], errors.messages['invalid unit'], unit))
                return def_unit

        if not reverse:
            # Преобразование вида: 1E3 -> кВ.
            self.unit_active_power = read_unit(self.unit_k_active_power, properties.unit_active_power,
                                               properties.unit_active_power[1])
            self.unit_reactive_power = read_unit(self.unit_k_reactive_power, properties.unit_reactive_power,
                                                 properties.unit_reactive_power[1])
            self.unit_full_power = read_unit(self.unit_k_full_power, properties.unit_full_power,
                                             properties.unit_full_power[1])
            self.unit_voltage = read_unit(self.unit_k_voltage, properties.unit_voltage,
                                          properties.unit_voltage[1])
            self.unit_current = read_unit(self.unit_k_current, properties.unit_current,
                                          properties.unit_current[1])
        else:
            # Преобразование вида: кВ -> 1E3.
            self.unit_k_active_power = read_unit(self.unit_active_power,
                                                 {unit: k for k, unit in properties.unit_active_power.iteritems()}, 1)
            self.unit_k_reactive_power = read_unit(self.unit_reactive_power,
                                                   {unit: k for k, unit in properties.unit_reactive_power.iteritems()},
                                                   1)
            self.unit_k_full_power = read_unit(self.unit_full_power,
                                               {unit: k for k, unit in properties.unit_full_power.iteritems()}, 1)
            self.unit_k_voltage = read_unit(self.unit_voltage,
                                            {unit: k for k, unit in properties.unit_voltage.iteritems()}, 1)
            self.unit_k_current = read_unit(self.unit_current,
                                            {unit: k for k, unit in properties.unit_current.iteritems()}, 1)

        messages.append(u'%s' % errors.messages['complete units conversion'])
        return errs, warnings, messages

    def rename_usel_in_vetvi(self, old_name, new_name):
        # Изменение имени узла во всех элементах.
        if old_name != new_name:
            for vetv in range(self.vetvi.child_count()):
                vetv = self.vetvi.child(vetv)
                input_parameters = vetv.input_parameters()
                if input_parameters.begin().value() == old_name:
                    input_parameters.begin().set_value(new_name)
                if input_parameters.end().value() == old_name:
                    input_parameters.end().set_value(new_name)
                if input_parameters.end_2().value() == old_name:
                    input_parameters.end_2().set_value(new_name)

    def set_actual_voltages(self):
        # Простановка реальных напряжений и углов поворота фаз на основе коэффициентов трансформации трансформаторов.
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Считывание базисного напряжения
        result = self.read_base_voltage(self.base_u_input)
        errs += result[0]
        warnings += result[1]
        messages += result[2]
        # Очистка реальных напряжений и углов поворота фаз в элементах.
        for element_name in self.vetvi.children_list():
            element = self.vetvi.find_child(element_name)
            input_parameters = element.input_parameters()
            input_parameters.real_voltage().set_value(0.)
            input_parameters.real_phase().set_value(0.)
        elements_list = self.vetvi.children_list()
        for element_name in elements_list:
            # Поиск трансформаторов в схеме.
            element = self.vetvi.find_child(element_name)
            input_parameters = element.input_parameters()
            if element.gui_parameters().view().value() in prog_settings.VETVI_VIEW_TRANSFORMERS:
                elements_list_1 = self.vetvi.children_list()
                real_voltage = 0.
                real_voltage_2 = 0.
                nominal_voltage = input_parameters.voltage().value()
                transformation_coefficient = input_parameters.transformation_coefficient().value()
                if transformation_coefficient:
                    real_voltage = nominal_voltage / transformation_coefficient
                transformation_coefficient_2 = input_parameters.transformation_coefficient_2().value()
                if transformation_coefficient_2:
                    real_voltage_2 = nominal_voltage / transformation_coefficient_2
                real_phase = input_parameters.transformation_coefficient_psi().value()
                real_phase_2 = input_parameters.transformation_coefficient_psi_2().value()
                try:
                    elements_list_1.remove(element_name)
                except ValueError:
                    pass
                if real_voltage:
                    # Если у трансформатора указан коэффициент трансформации, то у следующих за ним ветвей задается
                    # реальное напряжение и угол поворота фазы.
                    node = input_parameters.end().value()
                    if node and node != '0':
                        result = self.set_actual_voltage(elements_list_1[:], node, real_voltage, real_phase)
                        errs += result[0]
                        warnings += result[1]
                        messages += result[2]
                if real_voltage_2:
                    # Если у трансформатора указан коэффициент трансформации, то у следующих за ним ветвей задается
                    # реальное напряжение и угол поворота фазы.
                    node = input_parameters.end_2().value()
                    if node and node != '0':
                        result = self.set_actual_voltage(elements_list_1[:], node, real_voltage_2, real_phase_2)
                        errs += result[0]
                        warnings += result[1]
                        messages += result[2]
        return errs, warnings, messages

    def set_actual_voltage(self, elements_list, node, real_voltage, real_phase, set_real_voltage=True):
        # Функция являющаяся одной стадией прохода по элементам, следующим за данным элементом, которая изменяет
        # реальное напряжение и угол поворота фазы.
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        for element_name_1 in elements_list:
            elements_list_1 = elements_list[:]
            element_1 = self.vetvi.find_child(element_name_1)
            input_parameters_1 = element_1.input_parameters()
            begin_1 = input_parameters_1.begin().value()
            end_1 = input_parameters_1.end().value()
            end_2 = input_parameters_1.end_2().value()
            view_1 = element_1.gui_parameters().view().value()
            if not (view_1 in prog_settings.VETVI_VIEW_TRANSFORMERS and (end_1 == node or end_2 == node)):
                # Элемент учитывается, только если он не является трансформатором и при этом учитываемый узел
                # не является его концом. Т.е. нельзя заходить против направления трансформатора.
                if begin_1 == node or end_1 == node or end_2 == node:
                    # У элемента задаются реальное напряжение и (или) угол поворота фазы только если его начало или
                    # конец это учитываемый узел.
                    real_voltage_1 = input_parameters_1.real_voltage().value()
                    real_phase_1 = input_parameters_1.real_phase().value()
                    if real_voltage_1 and set_real_voltage:
                        # Если у элемента уже задано реальное напряжение и сказано, что нужно задавать реальное
                        # напряжение, то на этом элементе цикл заканчивается.
                        if real_voltage_1 != real_voltage:
                            if view_1 not in prog_settings.VETVI_VIEW_TRANSFORMERS:
                                # Если в узле node соединяются два элемента с разными напряжениями
                                # и этот элемент не трансформатор, то это ошибка.
                                # Если элемент является трансформатором, то на нем цикл заканчивается
                                errs.append(u'%s %s' % (errors.messages['different voltages converge'], node))
                    else:
                        # Если у элемента не задано реальное напряжение или оно задано, но сказано, что нужно
                        # изменять только фазу, то задаются реальное напряжение и (или) угол поворота фазы.
                        if set_real_voltage:
                            input_parameters_1.real_voltage().set_value(real_voltage)
                        # Угол поворота фазы прибавляется, т.к. другие трансформаторы тоже могут его изменять.
                        input_parameters_1.real_phase().set_value(real_phase_1 + real_phase)
                        next_node = ''
                        next_node_2 = ''
                        if begin_1 == node:
                            next_node = end_1
                            if end_2:
                                next_node_2 = end_2
                        elif end_1 == node:
                            next_node = begin_1
                        if next_node and next_node != '0':
                            set_real_voltage_1 = set_real_voltage
                            if view_1 in prog_settings.VETVI_VIEW_TRANSFORMERS:
                                # Если данный элемент является трансформатором, то у элементов за ним нужно
                                # изменять только угол поворота фазы.
                                set_real_voltage_1 = False
                            # Переход к следующему элементу. Данный элемент из списка удаляется, т.к. у него
                            # уже заданы данные реальное напряжение и угол поворота фазы.
                            try:
                                elements_list_1.remove(element_name_1)
                            except ValueError:
                                pass
                            self.set_actual_voltage(elements_list_1[:], next_node, real_voltage, real_phase,
                                                    set_real_voltage_1)
                            if next_node_2:
                                self.set_actual_voltage(elements_list_1[:], next_node_2, real_voltage, real_phase,
                                                        set_real_voltage_1)
        return errs, warnings, messages

    def three_phase_input_conversion(self):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        voltage_input = ['eds', 'voltage', 'voltage_addition']
        power_input = ['active_power', 'reactive_power']
        # Проход по всем элементам.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            # Проход по изменяемым параметрам.
            for param in voltage_input + power_input:
                if param in vetv.input_parameters().children_list():
                    if param in voltage_input:
                        try:
                            param = vetv.input_parameters().find_child(param)
                            param_value = param.value()
                            param.set_value(param_value / pow(3., 0.5))
                        except (ValueError, AttributeError, TypeError):
                            #errs.append(u'0 ОШИБКА: Неверно указан один из параметров элемента "%s"' % key).
                            pass
                    elif param in power_input:
                        try:
                            param = vetv.input_parameters().find_child(param)
                            param_value = param.value()
                            param.set_value(param_value / 3)
                        except (ValueError, AttributeError, TypeError):
                            #errs.append(u'0 ОШИБКА: Неверно указан один из параметров элемента "%s"' % key).
                            pass
        messages.append(u'%s' % errors.messages['complete input three-phase scheme'])
        return errs, warnings, messages

    def three_phase_output_conversion(self):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        voltage_output = ['u', 'u1', 'u2', 'real_u', 'real_e', 'real_u1', 'real_u2', 'real_u_usel']
        power_output = ['ist_s', 'priem_s', 's1', 's2', 's']
        # Преобразование полученных результатов к трехфазной симметричной схеме.
        # Умножить напряжения на 1,732, мощности на 3.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            vetv_output_parameters = vetv.output_parameters()
            if vetv_output_parameters.children_list():
                # Проход по изменяемым параметрам.
                for param in voltage_output + power_output:
                    if param in vetv_output_parameters.children_list():
                        if param in voltage_output:
                            param = vetv_output_parameters.find_child(param)
                            param.set_value(param.value() * pow(3., 0.5))
                        elif param in power_output:
                            param = vetv_output_parameters.find_child(param)
                            param.set_value(param.value() * 3)
        for usel in self.usli.children_list():
            usel = self.usli.find_child(usel)
            usel_output_parameters = usel.output_parameters()
            usel_output_parameters_list = usel_output_parameters.children_list()
            if usel_output_parameters_list:
                for param in voltage_output:
                    if param == 'u' and param in usel_output_parameters_list:
                        param = usel_output_parameters.find_child(param)
                        param.set_value(param.value() * pow(3., 0.5))
                    elif param == 'real_u_usel' and param in usel_output_parameters_list:
                        param = usel_output_parameters.find_child(param)
                        real_u_usel = []
                        for volt in param.value():
                            real_u_usel.append(volt * pow(3., 0.5))
                        param.set_value(real_u_usel)
        messages.append(u'%s' % errors.messages['complete output three-phase scheme'])
        return errs, warnings, messages

    def three_line_vetvi_input_conversion(self):
        errors = language.Errors()
        # Функция, которая создает в схеме три ветви, соединенные звездой вместо одного элемента, если в нем
        # заданы параметры двух других ветвей.
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Проход по всем элементам и создание дополнительных ветвей.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            if vetv.gui_parameters().view().value() in prog_settings.VETVI_VIEW_3_LINES:
                # Если графическое изображение элемента входит в список элементов, состоящих из трех ветвей.
                vetv_input_parameters = vetv.input_parameters()
                # Названия узлов, к которым присоединяются ветви.
                usel_2 = vetv_input_parameters.end().value()
                usel_3 = vetv_input_parameters.end_2().value()
                # Параметры дополнительных ветвей.
                real_voltage = vetv_input_parameters.real_voltage().value()
                real_phase = vetv_input_parameters.real_phase().value()
                active_resistance_2 = vetv_input_parameters.active_resistance_2().value()
                reactive_resistance_2 = vetv_input_parameters.reactive_resistance_2().value()
                active_resistance_3 = vetv_input_parameters.active_resistance_3().value()
                reactive_resistance_3 = vetv_input_parameters.reactive_resistance_3().value()
                active_conductivity_tr2 = vetv_input_parameters.active_conductivity_tr2().value()
                reactive_conductivity_tr2 = vetv_input_parameters.reactive_conductivity_tr2().value()
                # Создание нового узла (средней точки, соединяющей три ветви) - не требуется.
                usel_middle = u'%s_middle_node' % vetv.name()
                #usel_middle = classes.Usel(u'%s_middle_node' % vetv.name())
                #self.usli.add_child(usel_middle)
                # Конец ветви 1 - средняя точка.
                vetv_input_parameters.end().set_value(usel_middle)
                vetv_input_parameters.active_conductivity_tr2().set_value('')
                vetv_input_parameters.reactive_conductivity_tr2().set_value('')
                # Создание ветви 2.
                vetv_2 = classes.Vetv(u'%s_3_lines_2' % vetv.name())
                vetv_2_input_parameters = vetv_2.input_parameters()
                vetv_2_input_parameters.vetv_type().set_value('RX_post')
                vetv_2_input_parameters.real_voltage().set_value(real_voltage)
                vetv_2_input_parameters.real_phase().set_value(real_phase)
                vetv_2_input_parameters.active_resistance().set_value(active_resistance_2)
                vetv_2_input_parameters.reactive_resistance().set_value(reactive_resistance_2)
                vetv_2_input_parameters.active_conductivity_tr2().set_value(active_conductivity_tr2)
                vetv_2_input_parameters.reactive_conductivity_tr2().set_value(reactive_conductivity_tr2)
                vetv_2_input_parameters.begin().set_value(usel_middle)
                vetv_2_input_parameters.end().set_value(usel_2)
                self.vetvi.add_child(vetv_2)
                # Создание ветви 3.
                vetv_3 = classes.Vetv(u'%s_3_lines_3' % vetv.name())
                vetv_3_input_parameters = vetv_3.input_parameters()
                vetv_3_input_parameters.vetv_type().set_value('RX_post')
                vetv_3_input_parameters.real_voltage().set_value(real_voltage)
                vetv_3_input_parameters.real_phase().set_value(real_phase)
                vetv_3_input_parameters.active_resistance().set_value(active_resistance_3)
                vetv_3_input_parameters.reactive_resistance().set_value(reactive_resistance_3)
                vetv_3_input_parameters.active_conductivity_tr2().set_value(active_conductivity_tr2)
                vetv_3_input_parameters.reactive_conductivity_tr2().set_value(reactive_conductivity_tr2)
                vetv_3_input_parameters.begin().set_value(usel_middle)
                vetv_3_input_parameters.end().set_value(usel_3)
                self.vetvi.add_child(vetv_3)
        messages.append(u'%s' % errors.messages['complete input 3 line conversion'])
        return errs, warnings, messages

    def three_line_vetvi_output_conversion(self):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Проход по всем элементам и перенос результатов расчета из созданных дополнительных ветвей в обычные элементы.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            if vetv.gui_parameters().view().value() in prog_settings.VETVI_VIEW_3_LINES:
                # Если графическое изображение элемента входит в список элементов, состоящих из трех ветвей.
                vetv_output_parameters = vetv.output_parameters()
                vetv_2 = self.vetvi.find_child(u'%s_3_lines_2' % vetv.name())
                vetv_3 = self.vetvi.find_child(u'%s_3_lines_3' % vetv.name())
                if vetv_2 and vetv_3:
                    vetv_2_output_parameters = vetv_2.output_parameters()
                    vetv_3_output_parameters = vetv_3.output_parameters()
                    vetv_output_parameters.s1_2().set_value(vetv_2_output_parameters.s1().value())
                    vetv_output_parameters.s1_3().set_value(vetv_3_output_parameters.s1().value())
                    vetv_output_parameters.s2_2().set_value(vetv_2_output_parameters.s2().value())
                    vetv_output_parameters.s2_3().set_value(vetv_3_output_parameters.s2().value())
                    vetv_output_parameters.real_u_2().set_value(vetv_2_output_parameters.real_u().value())
                    vetv_output_parameters.real_u_3().set_value(vetv_3_output_parameters.real_u().value())
                    vetv_output_parameters.real_u1_2().set_value(vetv_2_output_parameters.real_u1().value())
                    vetv_output_parameters.real_u1_3().set_value(vetv_3_output_parameters.real_u1().value())
                    vetv_output_parameters.real_u2_2().set_value(vetv_2_output_parameters.real_u2().value())
                    vetv_output_parameters.real_u2_3().set_value(vetv_3_output_parameters.real_u2().value())
                    vetv_output_parameters.real_i_2().set_value(vetv_2_output_parameters.real_i().value())
                    vetv_output_parameters.real_i_3().set_value(vetv_3_output_parameters.real_i().value())
                    vetv_output_parameters.real_i2_tr().set_value(vetv_2_output_parameters.real_i2_tr().value())
                    vetv_output_parameters.real_i3_tr().set_value(vetv_3_output_parameters.real_i2_tr().value())
                    vetv_output_parameters.priem_s_2().set_value(vetv_2_output_parameters.priem_s().value())
                    vetv_output_parameters.priem_s_3().set_value(vetv_3_output_parameters.priem_s().value())
                    vetv_output_parameters.priem_s2_tr().set_value(vetv_2_output_parameters.priem_s2_tr().value())
                    vetv_output_parameters.priem_s3_tr().set_value(vetv_3_output_parameters.priem_s2_tr().value())
        messages.append(u'%s' % errors.messages['complete output 3 line conversion'])
        return errs, warnings, messages

    def transverse_vetvi_input_conversion(self):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Проход по всем элементам и создание поперечных ветвей.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            vetv_input_parameters = vetv.input_parameters()
            g1 = vetv_input_parameters.active_conductivity_tr1().value()
            b1 = vetv_input_parameters.reactive_conductivity_tr1().value()
            if g1 or b1:
                # Если у элемента задана активная или реактивная проводимость в начале.
                # Расчет сопротивлений поперечной ветви.
                z1 = 1 / (g1 + b1 * 1j)
                r1 = z1.real
                x1 = z1.imag
                # Создание поперечной ветви в начальном узле.
                vetv_new = classes.Vetv(u'%s_transverse_1' % vetv.name())
                vetv_new_input_parameters = vetv_new.input_parameters()
                # Установка параметров новой ветви.
                vetv_new_input_parameters.vetv_type().set_value('RX_post')
                vetv_new_input_parameters.real_voltage().set_value(vetv_input_parameters.real_voltage().value())
                vetv_new_input_parameters.real_phase().set_value(vetv_input_parameters.real_phase().value())
                vetv_new_input_parameters.begin().set_value(vetv_input_parameters.begin().value())
                vetv_new_input_parameters.end().set_value(u'0')
                vetv_new_input_parameters.active_resistance().set_value(r1)
                vetv_new_input_parameters.reactive_resistance().set_value(x1)
                self.vetvi.add_child(vetv_new)
            g2 = vetv_input_parameters.active_conductivity_tr2().value()
            b2 = vetv_input_parameters.reactive_conductivity_tr2().value()
            if g2 or b2:
                # Если у элемента задана активная или реактивная проводимость в конце.
                # Расчет сопротивлений поперечной ветви.
                z2 = 1 / (g2 + b2 * 1j)
                r2 = z2.real
                x2 = z2.imag
                # Создание поперечной ветви в начальном узле.
                vetv_new = classes.Vetv(u'%s_transverse_2' % vetv.name())
                vetv_new_input_parameters = vetv_new.input_parameters()
                # Установка параметров новой ветви.
                vetv_new_input_parameters.vetv_type().set_value('RX_post')
                vetv_new_input_parameters.real_voltage().set_value(vetv_input_parameters.real_voltage().value())
                vetv_new_input_parameters.real_phase().set_value(vetv_input_parameters.real_phase().value())
                vetv_new_input_parameters.begin().set_value(vetv_input_parameters.end().value())
                vetv_new_input_parameters.end().set_value(u'0')
                vetv_new_input_parameters.active_resistance().set_value(r2)
                vetv_new_input_parameters.reactive_resistance().set_value(x2)
                self.vetvi.add_child(vetv_new)
        messages.append(u'%s' % errors.messages['complete input transverse conversion'])
        return errs, warnings, messages

    def transverse_vetvi_output_conversion(self):
        errors = language.Errors()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []
        # Проход по всем элементам и перенос результатов расчета из поперечных ветвей в элементы.
        for vetv in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv)
            vetv_input_parameters = vetv.input_parameters()
            g1 = vetv_input_parameters.active_conductivity_tr1().value()
            b1 = vetv_input_parameters.reactive_conductivity_tr1().value()
            if g1 or b1:
                # Если у элемента задана активная или реактивная проводимость в начале.
                vetv_tr1 = self.vetvi.find_child(u'%s_transverse_1' % vetv.name())
                if vetv_tr1:
                    vetv_tr1_output_parameters = vetv_tr1.output_parameters()
                    vetv_output_parameters = vetv.output_parameters()
                    vetv_output_parameters.real_i1_tr().set_value(vetv_tr1_output_parameters.real_i().value())
                    vetv_output_parameters.priem_s1_tr().set_value(vetv_tr1_output_parameters.priem_s().value())
            g2 = vetv_input_parameters.active_conductivity_tr2().value()
            b2 = vetv_input_parameters.reactive_conductivity_tr2().value()
            if g2 or b2:
                # Если у элемента задана активная или реактивная проводимость в конце.
                vetv_tr2 = self.vetvi.find_child(u'%s_transverse_2' % vetv.name())
                if vetv_tr2:
                    vetv_tr2_output_parameters = vetv_tr2.output_parameters()
                    vetv_output_parameters = vetv.output_parameters()
                    vetv_output_parameters.real_i2_tr().set_value(vetv_tr2_output_parameters.real_i().value())
                    vetv_output_parameters.priem_s2_tr().set_value(vetv_tr2_output_parameters.priem_s().value())
        messages.append(u'%s' % errors.messages['complete output transverse conversion'])
        return errs, warnings, messages

    def calculate_power_loss(self):
        # Расчет суммарных потерь мощности в схеме. В расчет включаются мощности приемников всех элементов, кроме
        # элементов нагрузки, которые входят в состав списка VETVI_VIEW_LOAD в модуле prog_settings.
        sum_power_loss = 0. + 0.*1j
        for vetv_name in self.vetvi.children_list():
            vetv = self.vetvi.find_child(vetv_name)
            if vetv:
                view = vetv.gui_parameters().view().value()
                if view not in prog_settings.VETVI_VIEW_LOAD:
                    output_parameters = vetv.output_parameters()
                    sum_power_loss += output_parameters.priem_s().value()
                    sum_power_loss += output_parameters.priem_s1_tr().value()
                    sum_power_loss += output_parameters.priem_s2_tr().value()
                    if view in prog_settings.VETVI_VIEW_3_LINES:
                        sum_power_loss += output_parameters.priem_s_2().value()
                        sum_power_loss += output_parameters.priem_s_3().value()
                        sum_power_loss += output_parameters.priem_s3_tr().value()
        return sum_power_loss

    def calculate_plus(self, message_box=None, progress_bar=None, progress_bar_start=0, progress_bar_end=100,
                       set_actual_voltages=True):
        progress_bar_length = (progress_bar_end - progress_bar_start) / 100.
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []

        def message_box_append(message):
            if message_box is not None:
                for err in (message[0] + message[1] + message[2]):
                    message_box.append(err)

        def progress_bar_set_value(value):
            if progress_bar is not None:
                progress_bar.setValue(value * progress_bar_length + progress_bar_start)
        # Простановка реальных напряжений элементов.
        if set_actual_voltages:
            self.set_actual_voltages()
        # Создание копии схемы.
        node_scheme_copy = deep_copy_scheme(self, for_fast=True)
        progress_bar_set_value(10)

        if node_scheme_copy.scheme_type == '3_phase':
            # Если схема трехфазная симметричная, то нужно перевести ее в однофазную (без нейтрального провода).
            # Напряжения поделить на 1,732, а мощности на 3.
            # Реальное напряжение и базисное напряжение не трогаем.
            result = node_scheme_copy.three_phase_input_conversion()
            message_box_append(result)

        progress_bar_set_value(20)

        # Содание дополнительных ветвей из трехлучевых элементов.
        result = node_scheme_copy.three_line_vetvi_input_conversion()
        errs += result[0]
        warnings += result[1]
        messages += result[2]
        message_box_append(result)
        progress_bar_set_value(25)
        # Создание поперечных ветвей в скопированной схеме.
        result = node_scheme_copy.transverse_vetvi_input_conversion()
        errs += result[0]
        warnings += result[1]
        messages += result[2]
        message_box_append(result)
        progress_bar_set_value(30)

        # Чтение параметров элементов с учетом единиц измерения.
        result = node_scheme_copy.read_scheme(self.base_u_input)
        errs += result[0]
        warnings += result[1]
        messages += result[2]
        message_box_append(result)
        progress_bar_set_value(35)
        # Проверка на наличие критических ошибок после чтения параметров элементов.
        # Если критических ошибок нет, то выполняется расчет.
        if not errs:
            if progress_bar is not None:
                result = node_scheme_copy.calculate(progress_bar, 35 * progress_bar_length + progress_bar_start,
                                                    60 * progress_bar_length + progress_bar_start)
                errs += result[0]
                warnings += result[1]
                messages += result[2]
                message_box_append(result)
                progress_bar_set_value(60)
            else:
                result = node_scheme_copy.calculate(None, 0, 0)
                errs += result[0]
                warnings += result[1]
                messages += result[2]
                message_box_append(result)
            # Проверка на наличие критических ошибок после расчета.
            # Если критических ошибок нет, то выполняются дополнительные расчеты.
            if not errs:
                # Дополнительные расчеты + обратное преобразование единиц измерения.
                result = node_scheme_copy.calc_add()
                errs += result[0]
                warnings += result[1]
                messages += result[2]
                message_box_append(result)
                progress_bar_set_value(70)
                # Проверка результатов расчета.
                result = node_scheme_copy.results_check()
                errs += result[0]
                warnings += result[1]
                messages += result[2]
                message_box_append(result)
                progress_bar_set_value(75)

        if node_scheme_copy.scheme_type == '3_phase':
            # Обратное преобразование в трехфазную схему.
            result = node_scheme_copy.three_phase_output_conversion()
            errs += result[0]
            warnings += result[1]
            messages += result[2]
            message_box_append(result)
        progress_bar_set_value(80)
        # Перенос результатов расчетов из поперечных ветвей в обычные элементы.
        result = node_scheme_copy.transverse_vetvi_output_conversion()
        errs += result[0]
        warnings += result[1]
        messages += result[2]
        message_box_append(result)
        progress_bar_set_value(85)
        # Перенос результатов расчетов из, созданных дополнительных ветвей в обычные элементы.
        result = node_scheme_copy.three_line_vetvi_output_conversion()
        errs += result[0]
        warnings += result[1]
        messages += result[2]
        message_box_append(result)
        progress_bar_set_value(90)

        # Копирование выходных параметров из сохраненной копии списка элементов.
        for vetv in self.vetvi.children_list():
            vetv_copy = node_scheme_copy.vetvi.find_child(vetv)
            vetv = self.vetvi.find_child(vetv)
            if vetv_copy and vetv:
                vetv.remove_child('output_parameters')
                vetv.add_child(vetv_copy.output_parameters())
        progress_bar_set_value(95)
        # Копирование выходных параметров из сохраненной копии списка узлов.
        for usel in self.usli.children_list():
            usel_copy = node_scheme_copy.usli.find_child(usel)
            usel = self.usli.find_child(usel)
            if usel_copy and usel:
                usel.remove_child('output_parameters')
                usel.add_child(usel_copy.output_parameters())
        progress_bar_set_value(100)
        # Копирование количества итераций.
        self.n_shod = node_scheme_copy.n_shod

        return errs, warnings, messages
        
#-----Конец переопределенного класса схемы-----


import codecs
import cStringIO


class UnicodeWriter(object):
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """
    def __init__(self, f, delimiter=';', lineterminator='\n', encoding='utf-8'):
    # Redirect output to a queue
        self.queue = cStringIO.StringIO()
        self.writer = csv.writer(self.queue, delimiter=delimiter, lineterminator=lineterminator)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        self.writer.writerow([s.encode("utf-8") for s in row])
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)


def deep_copy_scheme(scheme=NodeScheme(), for_fast=False):
    # Создание полной независимой копии объекта схемы.
    if not isinstance(scheme, NodeScheme):
        return NodeScheme()

    new_scheme = NodeScheme(scheme.settings)
    new_scheme.base_u_input = scheme.base_u_input
    new_scheme.scheme_file_name = scheme.scheme_file_name
    new_scheme.scheme_type = scheme.scheme_type
    new_scheme.unit_active_power = scheme.unit_active_power
    new_scheme.unit_reactive_power = scheme.unit_reactive_power
    new_scheme.unit_full_power = scheme.unit_full_power
    new_scheme.unit_voltage = scheme.unit_voltage
    new_scheme.unit_current = scheme.unit_current
    new_scheme.print_power = scheme.print_power
    new_scheme.print_vetvi_power = scheme.print_vetvi_power
    new_scheme.print_vetvi_current = scheme.print_vetvi_current
    new_scheme.print_vetvi_voltage = scheme.print_vetvi_voltage
    new_scheme.print_usli_voltage = scheme.print_usli_voltage
    new_scheme.unit_k_active_power = scheme.unit_k_active_power
    new_scheme.unit_k_reactive_power = scheme.unit_k_reactive_power
    new_scheme.unit_k_full_power = scheme.unit_k_full_power
    new_scheme.unit_k_voltage = scheme.unit_k_voltage
    new_scheme.unit_k_current = scheme.unit_k_current
    new_scheme.base_u = scheme.base_u
    new_scheme.n_shod = scheme.n_shod
    new_scheme.use_temperature_coefficient = scheme.use_temperature_coefficient

    if for_fast:
        new_scheme.vetvi = classes.Parameters('vetvi')
    else:
        new_scheme.vetvi = classes.Parameters('vetvi')
    for vetv in scheme.vetvi.children_list():
        vetv = scheme.vetvi.find_child(vetv)
        if vetv:
            new_scheme.vetvi.add_child(classes.deep_copy_element(vetv))

    if for_fast:
        new_scheme.usli = classes.Parameters('usli')
    else:
        new_scheme.usli = classes.Parameters('usli')
    for usel in scheme.usli.children_list():
        usel = scheme.usli.find_child(usel)
        if usel:
            new_scheme.usli.add_child(classes.deep_copy_element(usel))

    return new_scheme


#-------Класс таблицы результатов расчета-------
class ResultsTable(object):

    def __init__(self):
        self.rows = []

    def writerow(self, row):
        self.rows.append(row)
#-------Конец класса таблицы результатов расчета-------


def write_powers(writer=ResultsTable(), node_scheme=NodeScheme(), round_n=3, file_point=','):
    lang = language.Other()
    # Запись таблицы результатов измерений (мощности приемников и источников).
    sum_priem_power = 0.
    sum_priem_power_tr = 0.
    writer.writerow([lang.table_results['consumers powers']])
    writer.writerow([lang.vetv_lang,
                     u'P, %s' % node_scheme.unit_active_power,
                     u'Q, %s' % node_scheme.unit_reactive_power,
                     u'S, %s' % node_scheme.unit_full_power,
                     u'cos(fi)',
                     u'P2, %s' % node_scheme.unit_active_power,
                     u'Q2, %s' % node_scheme.unit_reactive_power,
                     u'S2, %s' % node_scheme.unit_full_power,
                     u'cos(fi)2',
                     u'P3, %s' % node_scheme.unit_active_power,
                     u'Q3, %s' % node_scheme.unit_reactive_power,
                     u'S3, %s' % node_scheme.unit_full_power,
                     u'cos(fi)3',
                     u'P1 tr, %s' % node_scheme.unit_active_power,
                     u'Q1 tr, %s' % node_scheme.unit_reactive_power,
                     u'S1 tr, %s' % node_scheme.unit_full_power,
                     u'P2 tr, %s' % node_scheme.unit_active_power,
                     u'Q2 tr, %s' % node_scheme.unit_reactive_power,
                     u'S2 tr, %s' % node_scheme.unit_full_power,
                     u'P3 tr, %s' % node_scheme.unit_active_power,
                     u'Q3 tr, %s' % node_scheme.unit_reactive_power,
                     u'S3 tr, %s' % node_scheme.unit_full_power])
    for vetv in node_scheme.vetvi.children_list():
        vetv = node_scheme.vetvi.find_child(vetv)
        vetv_output_parameters = vetv.output_parameters()
        # Мощности в продольной ветви 1.
        p = round(vetv_output_parameters.priem_s().value().real, round_n)
        q = round(vetv_output_parameters.priem_s().value().imag, round_n)
        s_module = round(abs(vetv_output_parameters.priem_s().value()), round_n)
        # Мощности в продольной ветви 2.
        p2 = round(vetv_output_parameters.priem_s_2().value().real, round_n)
        q2 = round(vetv_output_parameters.priem_s_2().value().imag, round_n)
        s2_module = round(abs(vetv_output_parameters.priem_s_2().value()), round_n)
        # Мощности в продольной ветви 3.
        p3 = round(vetv_output_parameters.priem_s_3().value().real, round_n)
        q3 = round(vetv_output_parameters.priem_s_3().value().imag, round_n)
        s3_module = round(abs(vetv_output_parameters.priem_s_3().value()), round_n)
        # Мощности в поперечной ветви 1.
        p1_tr = round(vetv_output_parameters.priem_s1_tr().value().real, round_n)
        q1_tr = round(vetv_output_parameters.priem_s1_tr().value().imag, round_n)
        s1_module_tr = round(abs(vetv_output_parameters.priem_s1_tr().value()), round_n)
        # Мощности в поперечной ветви 2.
        p2_tr = round(vetv_output_parameters.priem_s2_tr().value().real, round_n)
        q2_tr = round(vetv_output_parameters.priem_s2_tr().value().imag, round_n)
        s2_module_tr = round(abs(vetv_output_parameters.priem_s2_tr().value()), round_n)
        # Мощности в поперечной ветви 3.
        p3_tr = round(vetv_output_parameters.priem_s3_tr().value().real, round_n)
        q3_tr = round(vetv_output_parameters.priem_s3_tr().value().imag, round_n)
        s3_module_tr = round(abs(vetv_output_parameters.priem_s3_tr().value()), round_n)
        # Расчет коэффициента мощности в продольной ветви 1.
        if abs(vetv_output_parameters.priem_s().value()) != 0:
            cos_fi = abs(round(vetv_output_parameters.priem_s().value().real /
                                abs(vetv_output_parameters.priem_s().value()), round_n))
        else:
            cos_fi = 0.0
        # Расчет коэффициента мощности в продольной ветви 2.
        if abs(vetv_output_parameters.priem_s_2().value()) != 0:
            cos_fi2 = abs(round(vetv_output_parameters.priem_s_2().value().real /
                                abs(vetv_output_parameters.priem_s_2().value()), round_n))
        else:
            cos_fi2 = 0.0
        # Расчет коэффициента мощности в продольной ветви 3.
        if abs(vetv_output_parameters.priem_s_3().value()) != 0:
            cos_fi3 = abs(round(vetv_output_parameters.priem_s_3().value().real /
                                abs(vetv_output_parameters.priem_s_3().value()), round_n))
        else:
            cos_fi3 = 0.0
        sum_priem_power += vetv_output_parameters.priem_s().value()
        sum_priem_power += vetv_output_parameters.priem_s_2().value()
        sum_priem_power += vetv_output_parameters.priem_s_3().value()
        sum_priem_power_tr += vetv_output_parameters.priem_s1_tr().value()
        sum_priem_power_tr += vetv_output_parameters.priem_s2_tr().value()
        sum_priem_power_tr += vetv_output_parameters.priem_s3_tr().value()
        writer.writerow([unicode(vetv.name()), unicode(p).replace('.', file_point),
                         unicode(q).replace('.', file_point), unicode(s_module).replace('.', file_point),
                         unicode(cos_fi).replace('.', file_point),
                         unicode(p2).replace('.', file_point), unicode(q2).replace('.', file_point),
                         unicode(s2_module).replace('.', file_point), unicode(cos_fi2).replace('.', file_point),
                         unicode(p3).replace('.', file_point), unicode(q3).replace('.', file_point),
                         unicode(s3_module).replace('.', file_point), unicode(cos_fi3).replace('.', file_point),
                         unicode(p1_tr).replace('.', file_point),
                         unicode(q1_tr).replace('.', file_point), unicode(s1_module_tr).replace('.', file_point),
                         unicode(p2_tr).replace('.', file_point), unicode(q2_tr).replace('.', file_point),
                         unicode(s2_module_tr).replace('.', file_point),
                         unicode(p3_tr).replace('.', file_point),
                         unicode(q3_tr).replace('.', file_point), unicode(s3_module_tr).replace('.', file_point)])
    # Расчет суммарных мощностей в продольных ветвях.
    p = round(sum_priem_power.real, round_n)
    q = round(sum_priem_power.imag, round_n)
    s_module = round(abs(sum_priem_power), round_n)
    # Расчет суммарных мощностей в начальных поперечных ветвях.
    p_tr = round(sum_priem_power_tr.real, round_n)
    q_tr = round(sum_priem_power_tr.imag, round_n)
    s_module_tr = round(abs(sum_priem_power_tr), round_n)
    # Расчет коэффициента суммарной мощности в продольных ветвях.
    if abs(sum_priem_power) != 0:
        cos_fi = abs(round(sum_priem_power.real / abs(sum_priem_power), round_n))
    else:
        cos_fi = 0.0
    writer.writerow([lang.table_results['total'], unicode(p).replace('.', file_point),
                     unicode(q).replace('.', file_point), unicode(s_module).replace('.', file_point),
                     unicode(cos_fi).replace('.', file_point), '', '', '', '', '', '', '', '',
                     unicode(p_tr).replace('.', file_point), unicode(q_tr).replace('.', file_point),
                     unicode(s_module_tr).replace('.', file_point), '', '', ''])

    sum_ist_power = 0.
    writer.writerow([lang.table_results['generators powers']])
    writer.writerow([lang.vetv_lang, u'P, %s' % node_scheme.unit_active_power,
                     u'Q, %s' % node_scheme.unit_reactive_power,
                     u'S, %s' % node_scheme.unit_full_power, u'cos(fi)'])
    for vetv in node_scheme.vetvi.children_list():
        vetv = node_scheme.vetvi.find_child(vetv)
        vetv_output_parameters = vetv.output_parameters()
        p = round(vetv_output_parameters.ist_s().value().real, round_n)
        q = round(vetv_output_parameters.ist_s().value().imag, round_n)
        s_module = round(abs(vetv_output_parameters.ist_s().value()), round_n)
        if abs(vetv_output_parameters.ist_s().value()) != 0:
            cos_fi = abs(round(vetv_output_parameters.ist_s().value().real /
                               abs(vetv_output_parameters.ist_s().value()), round_n))
        else:
            cos_fi = 0.0
        sum_ist_power += vetv_output_parameters.ist_s().value()
        writer.writerow([unicode(vetv.name()), unicode(p).replace('.', file_point),
                         unicode(q).replace('.', file_point), unicode(s_module).replace('.', file_point),
                         unicode(cos_fi).replace('.', file_point)])
    p = round(sum_ist_power.real, round_n)
    q = round(sum_ist_power.imag, round_n)
    s_module = round(abs(sum_ist_power), round_n)
    if abs(sum_ist_power) != 0:
        cos_fi = abs(round(sum_ist_power.real / abs(sum_ist_power), round_n))
    else:
        cos_fi = 0.0
    writer.writerow([lang.table_results['total'], unicode(p).replace('.', file_point),
                     unicode(q).replace('.', file_point), unicode(s_module).replace('.', file_point),
                     unicode(cos_fi).replace('.', file_point)])


def write_power_flows(writer=ResultsTable(), node_scheme=NodeScheme(), round_n=3, file_point=','):
    lang = language.Other()
    # Запись таблицы результатов измерений (потоки мощности элементов).
    writer.writerow([lang.table_results['power flows']])
    writer.writerow([lang.vetv_lang,
                     u'%s P, %s' % (lang.table_results['begin'], node_scheme.unit_active_power),
                     u'%s Q, %s' % (lang.table_results['begin'], node_scheme.unit_reactive_power),
                     u'%s S, %s' % (lang.table_results['begin'], node_scheme.unit_full_power),
                     u'%s cos(fi)' % lang.table_results['begin'],
                     u'%s P, %s' % (lang.table_results['end'], node_scheme.unit_active_power),
                     u'%s Q, %s' % (lang.table_results['end'], node_scheme.unit_reactive_power),
                     u'%s S, %s' % (lang.table_results['end'], node_scheme.unit_full_power),
                     u'%s cos(fi)' % lang.table_results['end'],
                     u'%s P2, %s' % (lang.table_results['begin'], node_scheme.unit_active_power),
                     u'%s Q2, %s' % (lang.table_results['begin'], node_scheme.unit_reactive_power),
                     u'%s S2, %s' % (lang.table_results['begin'], node_scheme.unit_full_power),
                     u'%s cos(fi)2' % lang.table_results['begin'],
                     u'%s P2, %s' % (lang.table_results['end'], node_scheme.unit_active_power),
                     u'%s Q2, %s' % (lang.table_results['end'], node_scheme.unit_reactive_power),
                     u'%s S2, %s' % (lang.table_results['end'], node_scheme.unit_full_power),
                     u'%s cos(fi)2' % lang.table_results['end'],
                     u'%s P3, %s' % (lang.table_results['begin'], node_scheme.unit_active_power),
                     u'%s Q3, %s' % (lang.table_results['begin'], node_scheme.unit_reactive_power),
                     u'%s S3, %s' % (lang.table_results['begin'], node_scheme.unit_full_power),
                     u'%s cos(fi)3' % lang.table_results['begin'],
                     u'%s P3, %s' % (lang.table_results['end'], node_scheme.unit_active_power),
                     u'%s Q3, %s' % (lang.table_results['end'], node_scheme.unit_reactive_power),
                     u'%s S3, %s' % (lang.table_results['end'], node_scheme.unit_full_power),
                     u'%s cos(fi)3' % lang.table_results['end']])
    for vetv in node_scheme.vetvi.children_list():
        vetv = node_scheme.vetvi.find_child(vetv)
        vetv_output_parameters = vetv.output_parameters()
        # Продольная ветвь 1. Начало.
        p_1_begin = round(vetv_output_parameters.s1().value().real, round_n)
        q_1_begin = round(vetv_output_parameters.s1().value().imag, round_n)
        s_1_begin_module = round(abs(vetv_output_parameters.s1().value()), round_n)
        if abs(vetv_output_parameters.s1().value()) != 0:
            cos_fi_1_begin = abs(round(vetv_output_parameters.s1().value().real /
                                       abs(vetv_output_parameters.s1().value()), round_n))
        else:
            cos_fi_1_begin = 0.0
        # Продольная ветвь 1. Конец.
        p_1_end = round(vetv_output_parameters.s2().value().real, round_n)
        q_1_end = round(vetv_output_parameters.s2().value().imag, round_n)
        s_1_end_module = round(abs(vetv_output_parameters.s2().value()), round_n)
        if abs(vetv_output_parameters.s2().value()) != 0:
            cos_fi_1_end = abs(round(vetv_output_parameters.s2().value().real /
                                     abs(vetv_output_parameters.s2().value()), round_n))
        else:
            cos_fi_1_end = 0.0
        # Продольная ветвь 2. Начало.
        p_2_begin = round(vetv_output_parameters.s1_2().value().real, round_n)
        q_2_begin = round(vetv_output_parameters.s1_2().value().imag, round_n)
        s_2_begin_module = round(abs(vetv_output_parameters.s1_2().value()), round_n)
        if abs(vetv_output_parameters.s1_2().value()) != 0:
            cos_fi_2_begin = abs(round(vetv_output_parameters.s1_2().value().real /
                                       abs(vetv_output_parameters.s1_2().value()), round_n))
        else:
            cos_fi_2_begin = 0.0
        # Продольная ветвь 2. Конец.
        p_2_end = round(vetv_output_parameters.s2_2().value().real, round_n)
        q_2_end = round(vetv_output_parameters.s2_2().value().imag, round_n)
        s_2_end_module = round(abs(vetv_output_parameters.s2_2().value()), round_n)
        if abs(vetv_output_parameters.s2_2().value()) != 0:
            cos_fi_2_end = abs(round(vetv_output_parameters.s2_2().value().real /
                                     abs(vetv_output_parameters.s2_2().value()), round_n))
        else:
            cos_fi_2_end = 0.0
        # Продольная ветвь 3. Начало.
        p_3_begin = round(vetv_output_parameters.s1_3().value().real, round_n)
        q_3_begin = round(vetv_output_parameters.s1_3().value().imag, round_n)
        s_3_begin_module = round(abs(vetv_output_parameters.s1_3().value()), round_n)
        if abs(vetv_output_parameters.s1_3().value()) != 0:
            cos_fi_3_begin = abs(round(vetv_output_parameters.s1_3().value().real /
                                       abs(vetv_output_parameters.s1_3().value()), round_n))
        else:
            cos_fi_3_begin = 0.0
        # Продольная ветвь 3. Конец.
        p_3_end = round(vetv_output_parameters.s2_3().value().real, round_n)
        q_3_end = round(vetv_output_parameters.s2_3().value().imag, round_n)
        s_3_end_module = round(abs(vetv_output_parameters.s2_3().value()), round_n)
        if abs(vetv_output_parameters.s2_3().value()) != 0:
            cos_fi_3_end = abs(round(vetv_output_parameters.s2_3().value().real /
                                     abs(vetv_output_parameters.s2_3().value()), round_n))
        else:
            cos_fi_3_end = 0.0

        writer.writerow([unicode(vetv.name()), unicode(p_1_begin).replace('.', file_point),
                         unicode(q_1_begin).replace('.', file_point),
                         unicode(s_1_begin_module).replace('.', file_point),
                         unicode(cos_fi_1_begin).replace('.', file_point), unicode(p_1_end).replace('.', file_point),
                         unicode(q_1_end).replace('.', file_point), unicode(s_1_end_module).replace('.', file_point),
                         unicode(cos_fi_1_end).replace('.', file_point),
                         unicode(p_2_begin).replace('.', file_point),
                         unicode(q_2_begin).replace('.', file_point),
                         unicode(s_2_begin_module).replace('.', file_point),
                         unicode(cos_fi_2_begin).replace('.', file_point), unicode(p_2_end).replace('.', file_point),
                         unicode(q_2_end).replace('.', file_point), unicode(s_2_end_module).replace('.', file_point),
                         unicode(cos_fi_2_end).replace('.', file_point),
                         unicode(p_3_begin).replace('.', file_point),
                         unicode(q_3_begin).replace('.', file_point),
                         unicode(s_3_begin_module).replace('.', file_point),
                         unicode(cos_fi_3_begin).replace('.', file_point), unicode(p_3_end).replace('.', file_point),
                         unicode(q_3_end).replace('.', file_point), unicode(s_3_end_module).replace('.', file_point),
                         unicode(cos_fi_3_end).replace('.', file_point)])


def write_currents(writer=ResultsTable(), node_scheme=NodeScheme(), round_n=3, file_point=','):
    lang = language.Other()
    # Запись таблицы результатов измерений (токи).
    writer.writerow([lang.table_results['currents']])
    writer.writerow([lang.vetv_lang, u'%s I, %s' % (lang.table_results['module'], node_scheme.unit_current),
                     u'%s I, °' % lang.table_results['phase'],
                     u'%s I2, %s' % (lang.table_results['module'], node_scheme.unit_current),
                     u'%s I2, °' % lang.table_results['phase'],
                     u'%s I3, %s' % (lang.table_results['module'], node_scheme.unit_current),
                     u'%s I3, °' % lang.table_results['phase'],
                     u'%s tr1, %s' % (lang.table_results['module'], node_scheme.unit_current),
                     u'%s tr1, °' % lang.table_results['phase'],
                     u'%s tr2, %s' % (lang.table_results['module'], node_scheme.unit_current),
                     u'%s tr2, °' % lang.table_results['phase'],
                     u'%s tr3, %s' % (lang.table_results['module'], node_scheme.unit_current),
                     u'%s tr3, °' % lang.table_results['phase']])
    for vetv in node_scheme.vetvi.children_list():
        vetv = node_scheme.vetvi.find_child(vetv)
        vetv_output_parameters = vetv.output_parameters()
        # Модуль и фаза тока в продольлной ветви.
        i_module = round(abs(vetv_output_parameters.real_i().value()), round_n)
        i_psi = round(math.degrees(cmath.phase(vetv_output_parameters.real_i().value())), round_n)
        if abs(i_psi) == 180:
            i_psi = 0.
            i_module *= -1
        # Модуль и фаза тока в продольлной ветви 2.
        i_2_module = round(abs(vetv_output_parameters.real_i_2().value()), round_n)
        i_2_psi = round(math.degrees(cmath.phase(vetv_output_parameters.real_i_2().value())), round_n)
        if abs(i_2_psi) == 180:
            i_2_psi = 0.
            i_2_module *= -1
        # Модуль и фаза тока в продольлной ветви 3.
        i_3_module = round(abs(vetv_output_parameters.real_i_3().value()), round_n)
        i_3_psi = round(math.degrees(cmath.phase(vetv_output_parameters.real_i_3().value())), round_n)
        if abs(i_3_psi) == 180:
            i_3_psi = 0.
            i_3_module *= -1
        # Модуль и фаза тока в поперечной ветви 1.
        i_module_tr1 = round(abs(vetv_output_parameters.real_i1_tr().value()), round_n)
        i_psi_tr1 = round(math.degrees(cmath.phase(vetv_output_parameters.real_i1_tr().value())), round_n)
        if abs(i_psi_tr1) == 180:
            i_psi_tr1 = 0.
            i_module_tr1 *= -1
        # Модуль и фаза тока в поперечной ветви 2.
        i_module_tr2 = round(abs(vetv_output_parameters.real_i2_tr().value()), round_n)
        i_psi_tr2 = round(math.degrees(cmath.phase(vetv_output_parameters.real_i2_tr().value())), round_n)
        if abs(i_psi_tr2) == 180:
            i_psi_tr2 = 0.
            i_module_tr2 *= -1
        # Модуль и фаза тока в поперечной ветви 3.
        i_module_tr3 = round(abs(vetv_output_parameters.real_i3_tr().value()), round_n)
        i_psi_tr3 = round(math.degrees(cmath.phase(vetv_output_parameters.real_i3_tr().value())), round_n)
        if abs(i_psi_tr3) == 180:
            i_psi_tr3 = 0.
            i_module_tr3 *= -1
        writer.writerow([unicode(vetv.name()), unicode(i_module).replace('.', file_point),
                         unicode(i_psi).replace('.', file_point),
                         unicode(i_2_module).replace('.', file_point), unicode(i_2_psi).replace('.', file_point),
                         unicode(i_3_module).replace('.', file_point), unicode(i_3_psi).replace('.', file_point),
                         unicode(i_module_tr1).replace('.', file_point), unicode(i_psi_tr1).replace('.', file_point),
                         unicode(i_module_tr2).replace('.', file_point), unicode(i_psi_tr2).replace('.', file_point),
                         unicode(i_module_tr3).replace('.', file_point), unicode(i_psi_tr3).replace('.', file_point)])


def write_vetvi_voltages(writer=ResultsTable(), node_scheme=NodeScheme(), round_n=3, file_point=','):
    lang = language.Other()
    # Запись таблицы результатов измерений (напряжения на элементах и ЭДС).
    writer.writerow([lang.table_results['elements voltages']])
    writer.writerow([lang.vetv_lang, u'%s U, %s' % (lang.table_results['module'], node_scheme.unit_voltage),
                     u'%s U, °' % lang.table_results['phase'],
                     u'%s U2, %s' % (lang.table_results['module'], node_scheme.unit_voltage),
                     u'%s U2, °' % lang.table_results['phase'],
                     u'%s U3, %s' % (lang.table_results['module'], node_scheme.unit_voltage),
                     u'%s U3, °' % lang.table_results['phase'],
                     u'%s %s, %s' % (lang.table_results['emf'], lang.table_results['module'], node_scheme.unit_voltage),
                     u'%s %s, °' % (lang.table_results['emf'], lang.table_results['phase'])])
    for vetv in node_scheme.vetvi.children_list():
        vetv = node_scheme.vetvi.find_child(vetv)
        vetv_output_parameters = vetv.output_parameters()
        u_module = round(abs(vetv_output_parameters.real_u().value()), round_n)
        u_psi = round(math.degrees(cmath.phase(vetv_output_parameters.real_u().value())), round_n)
        if abs(u_psi) == 180:
            u_psi = 0.
            u_module *= -1
        u_2_module = round(abs(vetv_output_parameters.real_u_2().value()), round_n)
        u_2_psi = round(math.degrees(cmath.phase(vetv_output_parameters.real_u_2().value())), round_n)
        if abs(u_2_psi) == 180:
            u_2_psi = 0.
            u_2_module *= -1
        u_3_module = round(abs(vetv_output_parameters.real_u_3().value()), round_n)
        u_3_psi = round(math.degrees(cmath.phase(vetv_output_parameters.real_u_3().value())), round_n)
        if abs(u_3_psi) == 180:
            u_3_psi = 0.
            u_3_module *= -1
        if vetv.input_parameters().vetv_type().value() in (prog_settings.EDS + prog_settings.NAGR_EDS):
            e_module = round(abs(vetv_output_parameters.real_e().value()), round_n)
            e_psi = round(math.degrees(cmath.phase(vetv_output_parameters.real_e().value())), round_n)
            if abs(e_psi) == 180:
                e_psi = 0
                e_module *= -1
        else:
            e_module = 0.
            e_psi = 0.
        writer.writerow([unicode(vetv.name()), unicode(u_module).replace('.', file_point),
                         unicode(u_psi).replace('.', file_point),
                         unicode(u_2_module).replace('.', file_point), unicode(u_2_psi).replace('.', file_point),
                         unicode(u_3_module).replace('.', file_point), unicode(u_3_psi).replace('.', file_point),
                         unicode(e_module).replace('.', file_point), unicode(e_psi).replace('.', file_point)])


def write_usli_voltages(writer=ResultsTable(), node_scheme=NodeScheme(), round_n=3, file_point=','):
    lang = language.Other()
    # Запись таблицы результатов измерений (напряжения в узлах).
    writer.writerow([lang.table_results['nodes voltages']])
    writer.writerow([lang.usel_lang, u'%s, %s' % (lang.table_results['module'], node_scheme.unit_voltage),
                     u'%s, °' % lang.table_results['phase']])
    for usel in node_scheme.usli.children_list():
        usel = node_scheme.usli.find_child(usel)
        usel_output_parameters = usel.output_parameters()
        if not usel_output_parameters.real_u_usel().value():
            writer.writerow([unicode(usel.name()), unicode(0.).replace('.', file_point),
                             unicode(0.).replace('.', file_point)])
        else:
            for u in usel_output_parameters.real_u_usel().value():
                u_module = round(abs(u), round_n)
                u_psi = round(math.degrees(cmath.phase(u)), round_n)
                if abs(u_psi) == 180:
                    u_psi = 0.
                    u_module *= -1
                writer.writerow([unicode(usel.name()), unicode(u_module).replace('.', file_point),
                                 unicode(u_psi).replace('.', file_point)])


def results_data(node_scheme=NodeScheme(), file_name=u'', round_n=3, file_point=','):
    errors = language.Errors()
    # Списки для записи ошибок, предупреждений и сообщений.
    errs = []
    warnings = []
    messages = []
    # Функция вывода данных в файл SCSV (Excel) или в объект с таблицей (класс results_table)
    # Если задано имя файла, то производится вывод данных в файл
    if file_name:
        # Попытка открыть файл для записи
        try:
            writer = UnicodeWriter(open(file_name, 'wb'), delimiter=';', lineterminator='\n', encoding='cp1251')
        except IOError:
            errs.append(u'%s: %s %s' % (errors.messages['error'], errors.messages['unable to write file'], file_name))
            return errs, warnings, messages
    # Если имя файла не задано, то производится запись в класс таблицы результатов расчета.
    else:
        writer = ResultsTable()

    # Вывод мощностей.
    if node_scheme.print_power:
        write_powers(writer, node_scheme, round_n, file_point)
        
    # Вывод потоков мощности в началах и концах элементов.
    if node_scheme.print_vetvi_power:
        write_power_flows(writer, node_scheme, round_n, file_point)
    
    # Вывод токов.
    if node_scheme.print_vetvi_current:
        write_currents(writer, node_scheme, round_n, file_point)
    
    # Вывод напряжений на элементах и ЭДС.
    if node_scheme.print_vetvi_voltage:
        write_vetvi_voltages(writer, node_scheme, round_n, file_point)
        
    # Вывод напряжений в узлах.
    if node_scheme.print_usli_voltage:
        write_usli_voltages(writer, node_scheme, round_n, file_point)
    
    if not file_name:
        return errs, warnings, messages, writer
    else:
        return errs, warnings, messages, True
