# -*- coding: utf-8 -*-
"""Модуль оптимизации схемы электроснабжения.
"""

# TODO: Добавить возможность создания исключений.

import language
import node_volt_calc_plus
import prog_settings


class SchemeOptimization(object):

    def __init__(self, node_scheme):
        # Объект оригинальной схемы.
        self._original_node_scheme = None
        # Карты состояний оригинальной схемы коммутационных аппаратов и источников,
        # которые используются для оптимизации.
        self._original_maps_scheme = ([], [])
        # Имена коммутационных аппаратов, которые используются для оптимизации.
        self._switchers_names = []
        # Имена элементов, у которых нужно контролировать направление потока мощности.
        self._elements_check_direction_names = []
        # Имена источников, которые используются для оптимизации.
        self._generators_names = []
        # Имена узлов, у которых нужно контролировать модуль напряжения.
        self._usli_names = []
        # Установка оригинальной схемы.
        self.set_node_scheme(node_scheme)
        # Списки, которые заполняются после завершения оптимизации.
        # Список схем с нормальной работой источников.
        self.schemes_normal = []
        # Список схем с перегрузкой хотя бы одного источника.
        self.schemes_overdrive = []
        # Список "плохих" схем. Плохие схемы - схемы, у которых в хотя бы одном элементе не прошла проверка
        # направления потока мощности или превышена перегрузка хотя бы одного источника.
        self.schemes_bad = []
        # Количество рассчитанных схем.
        self.number_schemes = 0

    def set_node_scheme(self, node_scheme):
        # Установка оригинальной схемы.
        if isinstance(node_scheme, node_volt_calc_plus.NodeScheme):
            self._original_node_scheme = node_volt_calc_plus.deep_copy_scheme(node_scheme)
            elements_names = self.elements_names()
            self._switchers_names = elements_names[0]
            self._elements_check_direction_names = elements_names[1]
            self._generators_names = elements_names[2]
            self._usli_names = self.usli_names()
            self.create_original_maps_scheme()
            self.number_schemes = 0
            self.schemes_normal = []
            self.schemes_overdrive = []
            self.schemes_bad = []

    def create_original_maps_scheme(self):
        # Создание карт состояний коммутационных аппаратов и источников оригинальной схемы.
        original_switchers_map = []
        original_generators_map = []
        # Проход по всем коммутационным аппаратам и определение их состояния.
        for switcher_name in self._switchers_names:
            switcher = self._original_node_scheme.vetvi.find_child(switcher_name)
            if switcher:
                input_parameters = switcher.input_parameters()
                active_resistance = input_parameters.active_resistance().value()
                reactive_resistance = input_parameters.reactive_resistance().value()
                if active_resistance >= 0. and reactive_resistance >= 0.:
                    original_switchers_map.append(True)
                else:
                    original_switchers_map.append(False)
            else:
                original_switchers_map = []
                break
        # Проход по всем источникам и определение их состояния.
        for generator_name in self._generators_names:
            generator = self._original_node_scheme.vetvi.find_child(generator_name)
            if generator:
                generator_switcher_name = generator.optimization_parameters().generator_switcher().value()
                if generator_switcher_name:
                    generator_switcher = self._original_node_scheme.vetvi.find_child(generator_switcher_name)
                    if generator_switcher:
                        input_parameters = generator_switcher.input_parameters()
                        active_resistance = input_parameters.active_resistance().value()
                        reactive_resistance = input_parameters.reactive_resistance().value()
                        if active_resistance >= 0. and reactive_resistance >= 0.:
                            original_generators_map.append(True)
                        else:
                            original_generators_map.append(False)
                    else:
                        original_generators_map = []
                        break
                else:
                    original_generators_map = []
                    break
        self._original_maps_scheme = (original_switchers_map, original_generators_map)

    def original_node_scheme(self):
        return self._original_node_scheme

    def elements_names(self):
        # Список имен коммутационных аппаратов, используемых в оптимизации схемы.
        switchers_names = []
        # Список имен элементов, в которых контролируется направление потока мощности.
        elements_check_direction_names = []
        # Список имен генераторов, используемых в оптимизации схемы, отсортированный по приоритету источника.
        generators_names_sorted = []
        if not isinstance(self._original_node_scheme, node_volt_calc_plus.NodeScheme):
            return switchers_names, elements_check_direction_names, generators_names_sorted

        generators_names = []
        for vetv_name in self._original_node_scheme.vetvi.children_list():
            # Поиск элементов (коммутационных аппаратов), используемых в оптимизации схемы.
            # Поиск источников, используемых в оптимизации схемы.
            vetv = self._original_node_scheme.vetvi.find_child(vetv_name)
            if vetv:
                use_for_optimization = vetv.optimization_parameters().use_for_optimization().value()
                switcher_permit = vetv.optimization_parameters().switcher_permit().value()
                view = vetv.gui_parameters().view().value()
                if view in prog_settings.VETVI_VIEW_SWITCHERS:
                    # Если элемент относится к списку коммутационных аппаратов.
                    if use_for_optimization and switcher_permit:
                        # Если коммутационный аппарат используется для оптимизации и если можно изменять его состояние,
                        # то его имя добавляется к списку переключателей.
                        switchers_names.append(vetv_name)
                elif view in prog_settings.VETVI_VIEW_GENERATORS:
                    # Если элемент относится к списку источников.
                    if use_for_optimization:
                        # Если источник испльзуется для оптимизации, то его имя и приоритет добавляется к списку
                        # источников.
                        priority = vetv.optimization_parameters().generator_priority().value()
                        generators_names.append({'name': vetv.name(), 'priority': priority})
                if use_for_optimization:
                    if vetv.optimization_parameters().check_direction().value():
                        # Если элемент используется для оптимизации и необходимо контролировать направление потока
                        # мощности, то элемент добавляется к соответствующему списку.
                        elements_check_direction_names.append(vetv_name)
        # Сортировка списка источников по приоритету.
        generators_names.sort(key=lambda x: x['priority'])
        generators_names_sorted = [g['name'] for g in generators_names]
        return switchers_names, elements_check_direction_names, generators_names_sorted

    def usli_names(self):
        # Список узлов, в которых нужно проводить проверку напряжения.
        usli_check_names = []
        if not isinstance(self._original_node_scheme, node_volt_calc_plus.NodeScheme):
            return usli_check_names

        for usel_name in self._original_node_scheme.usli.children_list():
            usel = self._original_node_scheme.usli.find_child(usel_name)
            if usel:
                if usel.optimization_parameters().use_for_optimization().value():
                    usli_check_names.append(usel.name())
        return usli_check_names

    def switch_switcher(self, switcher, switch=True):
        active_resistance = switcher.input_parameters().active_resistance()
        reactive_resistance = switcher.input_parameters().reactive_resistance()
        buffer_active_resistance = switcher.buffer_parameters().active_resistance()
        buffer_reactive_resistance = switcher.buffer_parameters().reactive_resistance()
        if switch:
            # Если коммутационный аппарат должен быть включен, то устанавливаются
            # соответствующие сопротивления.
            if active_resistance.value() < 0. or reactive_resistance.value() < 0.:
                # Если коммутационный аппарат не включен, то устанавливаются сопротивления из его буфера.
                if buffer_active_resistance.value() >= 0. and buffer_reactive_resistance.value() >= 0.:
                    active_resistance.set_value(buffer_active_resistance.value())
                    reactive_resistance.set_value(buffer_reactive_resistance.value())
                else:
                    active_resistance.set_value(0.)
                    reactive_resistance.set_value(0.)
        else:
            # Если коммутационный аппарат должен быть выключен, то устанавливаются сопротивления,
            # равные -1, а старые сопротивления записываются в буфер.
            # Сохранение сопротивлений в буфер, если сопротивления больше 0.
            if active_resistance.value() >= 0. and reactive_resistance.value() >= 0.:
                buffer_reactive_resistance.set_value(active_resistance.value())
                buffer_active_resistance.set_value(reactive_resistance.value())
            active_resistance.set_value(-1.)
            reactive_resistance.set_value(-1.)

    def create_scheme(self, switchers_map, generators_map):
        # Создание схемы по "картам" источников и коммутационных аппаратов.
        # Карта - список состояний источников (их коммутационных аппаратов) или коммутационных аппаратов.
        # Количество состояний в картах должно совпадать с количеством источников или коммутационных аппаратов.
        if not isinstance(self._original_node_scheme, node_volt_calc_plus.NodeScheme):
            return False

        if len(self._switchers_names) != len(switchers_map):
            return False

        if len(self._generators_names) != len(generators_map):
            return False

        node_scheme = node_volt_calc_plus.deep_copy_scheme(self._original_node_scheme)
        # Выходные карты состояний, которые используются для сравнения схем и поиска одинаковых схем.
        output_switchers_map = switchers_map
        output_generators_map = generators_map

        # Проход по коммутационным аппаратам и изменение их состояния в соответствии с заданной картой.
        for i, switcher_name in enumerate(self._switchers_names):
            switcher = node_scheme.vetvi.find_child(switcher_name)
            if switcher:
                if switcher.optimization_parameters().switcher_permit().value():
                    # Если можно изменить состояние коммутационного аппарата.
                    switch = switchers_map[i]
                    self.switch_switcher(switcher, switch)
                else:
                    # Если нельзя изменить состояние коммутационного аппарата, то в выходную карту
                    # коммутационных аппаратов записывается его текущее состояние.
                    active_resistance = switcher.input_parameters().active_resistance()
                    reactive_resistance = switcher.input_parameters().reactive_resistance()
                    if active_resistance.value() >= 0. and reactive_resistance.value() >= 0.:
                        output_switchers_map[i] = True
                    else:
                        output_switchers_map[i] = False
        # Проход по источникам.
        for i, generator_name in enumerate(self._generators_names):
            generator = node_scheme.vetvi.find_child(generator_name)
            if generator:
                optimization_parameters = generator.optimization_parameters()
                generator_switcher_name = optimization_parameters.generator_switcher().value()
                if generator_switcher_name:
                    # Если задан коммутационный аппарат источника.
                    generator_switcher = node_scheme.vetvi.find_child(generator_switcher_name)
                    if optimization_parameters.switcher_permit().value():
                        # Если можно изменить состояние коммутационного аппарата источника.
                        if generator_switcher:
                            # Если найден коммутационный аппарат источника.
                            switch = generators_map[i]
                            self.switch_switcher(generator_switcher, switch)
                    else:
                        # Если нельзя изменить состояние коммутационного аппарата, то в выходную карту
                        # источников записывается его текущее состояние.
                        if generator_switcher:
                            active_resistance = generator_switcher.input_parameters().active_resistance()
                            reactive_resistance = generator_switcher.input_parameters().reactive_resistance()
                            if active_resistance.value() >= 0. and reactive_resistance.value() >= 0.:
                                output_generators_map[i] = True
                            else:
                                output_generators_map[i] = False

        return node_scheme, output_switchers_map, output_generators_map

    def calculate_number_switching(self, maps_scheme):
        # Определение количество переключений по картам схемы maps_scheme и оригинальным картам.
        number_switching = 0
        # Если количество элементов в картах схемы и оригинальных картах не совпадает, то функция завершается.
        if len(maps_scheme[0]) != len(self._original_maps_scheme[0]):
            return -1
        if len(maps_scheme[1]) != len(self._original_maps_scheme[1]):
            return -1
        # Сверка состояний коммутационных аппаратов с оригинальной картой.
        for i, state in enumerate(maps_scheme[0]):
            if state != self._original_maps_scheme[0][i]:
                number_switching += 1
        # Сверка состояний источников с оригинальной картой.
        for i, state in enumerate(maps_scheme[1]):
            if state != self._original_maps_scheme[1][i]:
                number_switching += 1
        return number_switching

    def check_list_schemes(self, list_maps_schemes, maps_scheme):
        # Проверка на наличие такой же схемы по выходным картам состояний в списке карт пройденных вариантов схем.
        for i_maps_scheme in list_maps_schemes:
            if i_maps_scheme[0] == maps_scheme[0] and i_maps_scheme[1] == maps_scheme[1]:
                # Если карты состояний коммутационных аппаратов и карты состояний источников
                # i-й схемы и проверяемой схемы совпадают, то функция возвращает True.
                return True
        return False

    def check_usli_voltages(self, node_scheme):
        # Проверка уровней напряжения в узлах схемы (минимальное и максимальное).
        result = True
        for usel_name in self._usli_names:
            usel = node_scheme.usli.find_child(usel_name)
            if usel:
                usel_voltage = usel.output_parameters().real_u_usel().value()
                if len(usel_voltage) > 0:
                    usel_voltage = abs(usel_voltage[0])
                    optimization_parameters = usel.optimization_parameters()
                    minimum_voltage = optimization_parameters.minimum_voltage().value()
                    maximum_voltage = optimization_parameters.maximum_voltage().value()
                    if usel_voltage < minimum_voltage or usel_voltage > maximum_voltage:
                        result = False
                elif len(usel_voltage) == 0:
                    result = False
        return result

    def check_next_generator_priority(self, g):
        if g < len(self._generators_names) - 1:
            generator_name = self._generators_names[g]
            next_generator_name = self._generators_names[g + 1]
            generator = self._original_node_scheme.vetvi.find_child(generator_name)
            next_generator = self._original_node_scheme.vetvi.find_child(next_generator_name)
            if generator and next_generator:
                generator_priority = generator.optimization_parameters().generator_priority().value()
                next_generator_priority = next_generator.optimization_parameters().generator_priority().value()
                if generator_priority == next_generator_priority:
                    return True
        return False

    def check_power_of_generators(self, node_scheme):
        # Проверка мощности источников. Если мощности источников меньше нормальных, то функция возвращает 0,
        # если больше нормальных, но меньше перегрузочных, то возвращает 1,
        # если больше перегрузочных, то возвращает 2.
        result = 0
        for generator_name in self._generators_names:
            generator = node_scheme.vetvi.find_child(generator_name)
            if generator:
                optimization_parameters = generator.optimization_parameters()
                power = generator.output_parameters().ist_s().value()
                if power:
                    nominal_active_power = optimization_parameters.generator_nominal_active_power().value()
                    nominal_reactive_power = optimization_parameters.generator_nominal_reactive_power().value()
                    nominal_full_power = optimization_parameters.generator_nominal_full_power().value()
                    normal_coefficient = optimization_parameters.generator_normal_power_coefficient().value()
                    overdrive_coefficient = optimization_parameters.generator_overdrive_power_coefficient().value()
                    active_range_1 = nominal_active_power * normal_coefficient
                    active_range_2 = nominal_active_power * overdrive_coefficient
                    reactive_range_1 = nominal_reactive_power * normal_coefficient
                    reactive_range_2 = nominal_reactive_power * overdrive_coefficient
                    full_range_1 = nominal_full_power * normal_coefficient
                    full_range_2 = nominal_full_power * overdrive_coefficient
                    result_active = 0
                    result_reactive = 0
                    result_full = 0
                    if active_range_1 and active_range_2:
                        if power.real > active_range_1 and power.real <= active_range_2:
                            result_active = 1
                        elif power.real > active_range_2:
                            result_active = 2
                    if reactive_range_1 and reactive_range_2:
                        if power.imag > reactive_range_1 and power.imag <= reactive_range_2:
                            result_reactive = 1
                        elif power.imag > reactive_range_2:
                            result_reactive = 2
                    if full_range_1 and full_range_2:
                        if abs(power) > full_range_1 and abs(power) <= full_range_2:
                            result_full = 1
                        elif abs(power) > full_range_2:
                            result_full = 2
                    result_1 = max(result_active, result_reactive, result_full)
                    result = max(result, result_1)
        return result

    def check_directions(self, node_scheme):
        # Проверка направлений потоков мощности в элементах.
        result = True
        for element_name in self._elements_check_direction_names:
            element = node_scheme.vetvi.find_child(element_name)
            if element:
                power_flow_1 = element.output_parameters().s1().value()
                if power_flow_1.real < 0. or power_flow_1.imag < 0.:
                    return False
        return result

    def check_scheme(self, node_scheme, scheme_number, maps_scheme):
        lang = language.Other()
        good_voltages_scheme = 0
        normal_scheme = 0
        overdrive_scheme = 0
        bad_scheme = 0
        # Название варианта схемы.
        if scheme_number:
            variant_name = u'%s %s' % (lang.optimization_table_rows['variant'], scheme_number)
        else:
            variant_name = lang.optimization_table_rows['original_scheme']
        # Определение количества перелкючений в схеме по сравнению с оригинальной.
        number_switching = self.calculate_number_switching(maps_scheme)
        # Проверка напряжений в контролируемых узлах.
        if self.check_usli_voltages(node_scheme):
            good_voltages_scheme = 1
            # Расчет суммарных потерь мощности в схеме.
            sum_power_loss = node_scheme.calculate_power_loss()
            # Проверка направлений потоков мощности в элементах.
            if self.check_directions(node_scheme):
                # Если проверка пройдена, то производится проверка мощностей источников.
                check_power_of_generators = self.check_power_of_generators(node_scheme)
                if check_power_of_generators == 0:
                    # Мощность всех источников меньше нормальной.
                    normal_scheme = 1
                    self.schemes_normal.append([variant_name, node_scheme, sum_power_loss.real, sum_power_loss.imag,
                                                number_switching])
                elif check_power_of_generators == 1:
                    # Мощность одного из источников больше нормальной, но меньше допустимой перегрузки.
                    overdrive_scheme = 1
                    self.schemes_overdrive.append([variant_name, node_scheme, sum_power_loss.real, sum_power_loss.imag,
                                                   number_switching])
                else:
                    # Мощность одного из источников больше допустимой перегрузки.
                    bad_scheme = 1
                    self.schemes_bad.append([variant_name, node_scheme, sum_power_loss.real, sum_power_loss.imag,
                                             number_switching])
            else:
                # Если не пройдена проверка по направлению потоков мощности в элементах, то схема считается "плохой".
                bad_scheme = 1
                self.schemes_bad.append([variant_name, node_scheme, sum_power_loss.real, sum_power_loss.imag,
                                         number_switching])
        return good_voltages_scheme, normal_scheme, overdrive_scheme, bad_scheme

    def optimize(self, message_box=None, progress_bar=None, progress_bar_start=0, progress_bar_end=100):
        progress_bar_length = (progress_bar_end - progress_bar_start) / 100.
        errors = language.Errors()
        messages_dict = language.Messages()
        other_lang = language.Other()
        # Списки для записи ошибок, предупреждений и сообщений.
        errs = []
        warnings = []
        messages = []

        def message_box_append(message):
            if message_box is not None:
                for mes in (message[0] + message[1] + message[2]):
                    message_box.append(mes)

        def progress_bar_set_value(value):
            if progress_bar is not None:
                progress_bar.setValue(value * progress_bar_length + progress_bar_start)

        # Список карт пройденных вариантов схем.
        list_maps_schemes = []
        # Очистка списков схем с нормальными режимами работы источников, схем с перегрузкой источников и
        # "плохих" схем.
        self.schemes_normal = []
        self.schemes_overdrive = []
        self.schemes_bad = []
        # Помещение оригинальной схемы в один из списков.
        self.check_scheme(self.original_node_scheme(), 0, self._original_maps_scheme)
        # Количество рассмотренных схем.
        number_schemes = 0
        number_calculated_schemes = 0
        number_good_voltages_schemes = 0
        number_normal_schemes = 0
        number_overdrive_schemes = 0
        number_bad_schemes = 0
        # Количество переключателей.
        number_switchers = len(self._switchers_names)
        # Количество источников.
        number_generators = len(self._generators_names)
        progress_bar_set_value(5)
        # Если количество используемых для оптимизации переключателей и источников равно нулю,
        # то оптимизация не производится.
        if number_generators == 0 and number_switchers == 0:
            error_text = u'%s: %s. %s' % (errors.messages['error'], errors.messages['can not done optimization'],
                                          errors.messages['not set generators and switchers'])
            errs.append(error_text)
            message_box_append(([error_text], [], []))
            return errs, warnings, messages
        message_text = u'%s: %s' % (messages_dict.messages['number switchers'], number_switchers)
        message_text += u'\n%s: %s' % (messages_dict.messages['number generators'], number_generators)
        messages.append(message_text)
        message_box_append(([message_text], [], []))
        # Строка состоящая из единиц, длина строки - количество переключателей.
        binar = ''.rjust(number_switchers, '1')
        if binar:
            n_operations = int(binar, 2)    # Количество операций - 1
        else:
            n_operations = 0
        progress_bar_set_value(10)
        # Цикл прохода по возможным состояниям переключателей.
        for switch_n in range(n_operations + 1):
            def my_bool(value):
                if value == '0':
                    value = ''
                return bool(value)
            if n_operations != 0:
                # Создание карты состояний коммутационных аппаратов.
                switchers_map_str = (bin(switch_n)[2:]).zfill(number_switchers)
                switchers_map = [my_bool(i) for i in switchers_map_str]
            else:
                switchers_map = []
            # Цикл подключения источников по приоритету.
            for g in range(-1, number_generators):
                a = 10 + (switch_n + 1) * 85 / (n_operations + 1)
                b = 10 + switch_n * 85 / n_operations
                # Создание карты состояний источников.
                if g > -1:
                    if self.check_next_generator_priority(g):
                        continue
                    generators_map = [True for i in range(number_generators)]
                    if g < number_generators - 1:
                        for i in range(g + 1, number_generators):
                            generators_map[i] = False
                else:
                    generators_map = [False for i in range(number_generators)]
                # Создание схемы на основе карт состояний коммутационных аппаратов и источников.
                scheme = self.create_scheme(switchers_map, generators_map)
                if scheme:
                    node_scheme = scheme[0]
                    maps_scheme = [scheme[1], scheme[2]]
                    # Проверка на наличие такой же схемы по выходным картам состояний в списке карт пройденных
                    # вариантов схем.
                    if not self.check_list_schemes(list_maps_schemes, maps_scheme):
                        # Если подобной схемы нет, то производится добавление карт схемы к списку и расчет схемы.
                        number_schemes += 1
                        list_maps_schemes.append(maps_scheme)
                        result = node_scheme.calculate_plus(None, None, 0, 0)
                        # Проверка на наличие критических ошибок после расчета.
                        if not result[0]:
                            number_calculated_schemes += 1
                            # Проверки схемы и добавление в один из списков.
                            checks = self.check_scheme(node_scheme, number_schemes, maps_scheme)
                            number_good_voltages_schemes += checks[0]
                            number_normal_schemes += checks[1]
                            number_overdrive_schemes += checks[2]
                            number_bad_schemes += checks[3]
                progress_bar_set_value(a + (g + 2) * (b - a) / (number_generators + 1))
            progress_bar_set_value(10 + (switch_n + 1) * 85 / (n_operations + 1))
        progress_bar_set_value(95)
        # Сортировка списков схем по потерям активной мощности, а затем по количеству переключений.
        # Потери активной мощности для сортировки округляются до 5 знака.
        self.schemes_normal = sorted(self.schemes_normal, key=lambda s: (round(s[2], 5), s[4]))
        self.schemes_overdrive = sorted(self.schemes_overdrive, key=lambda s: (round(s[2], 5), s[4]))

        message_text = u'%s: %s' % (messages_dict.messages['number schemes'], number_schemes)
        message_text += u'\n%s: %s' % (messages_dict.messages['number calculated schemes'], number_calculated_schemes)
        message_text += u'\n%s: %s' % (messages_dict.messages['number good voltages schemes'],
                                       number_good_voltages_schemes)
        message_text += u'\n%s: %s' % (messages_dict.messages['number normal schemes'],
                                       number_normal_schemes)
        message_text += u'\n%s: %s' % (messages_dict.messages['number overdrive schemes'],
                                       number_overdrive_schemes)
        messages.append(message_text)
        message_box_append(([message_text], [], []))
        progress_bar_set_value(100)
        return errs, warnings, messages